#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

__author__ = "John Eslick"

import os
import csv

import pyomo.environ as pyo
from pyomo.network import Arc
from pyomo.common.fileutils import this_file_dir

from idaes.core import FlowsheetBlock
from idaes.generic_models.properties.core.generic.generic_property import (
    GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
    GenericReactionParameterBlock)
import idaes.generic_models.unit_models as um # um = unit models
import idaes.core.util as iutil
import idaes.core.util.tables as tables
import idaes.core.util.scaling as iscale
import idaes.core.plugins
from idaes.power_generation.properties.natural_gas_PR import get_prop, get_rxn
from idaes.core.solvers import use_idaes_solver_configuration_defaults


def tag_model(m):
    """
    Create to dictionaries. One is called tags with tags for keys and expressions
    for values. The second is called tag_format with tags for keys and a
    formatting string.  The formating string controls the number format and can
    be used to add additional text, to report units for example. The two
    dictionaries are returned, and also attached to the model.

    Args:
        m: (ConcreteModel) model to tag

    Returns:
        (dict) tags, (dict) tag_format
    """
    tags = {} # dict of with tag keys and expressions for their values
    tag_format = {} # format string for the tags
    def new_tag(name, expr, format):
        # funcion to keep it more compact
        tags[name] = expr
        tag_format[name] = format
    # Create a dict with Arc name keys and state block values
    stream_states = tables.stream_states_dict(
        tables.arcs_to_stream_dict(m.fs, descend_into=False))
    for i, s in stream_states.items(): # create the tags for steam quantities
        new_tag(f"{i}_Fvol", expr=s.flow_vol, format="{:.1f} m^3/s")
        new_tag(f"{i}_Fmol", expr=s.flow_mol/1000, format="{:.3f} kmol/s")
        new_tag(f"{i}_F", expr=s.flow_mass, format="{:.3f} kg/s")
        new_tag(f"{i}_P", expr=s.pressure/1000, format="{:.3f} kPa")
        new_tag(f"{i}_T", expr=s.temperature, format="{:.2f} K")
        for c in s.mole_frac_comp:
            new_tag(f"{i}_y{c}", expr=s.mole_frac_comp[c]*100, format="{:.3f}%")
    # Tags for non-stream things
    new_tag(
        "cmp1_power",
        expr=m.fs.cmp1.control_volume.work[0]/1e6,
        format="{:.2f} MW")
    new_tag(
        "gts1_power",
        expr=m.fs.gts1.control_volume.work[0]/1e6,
        format="{:.2f} MW")
    new_tag(
        "gts2_power",
        expr=m.fs.gts2.control_volume.work[0]/1e6,
        format="{:.2f} MW")
    new_tag(
        "gts3_power",
        expr=m.fs.gts3.control_volume.work[0]/1e6,
        format="{:.2f} MW")
    new_tag(
        "cmp1_head_isen",
        expr=m.fs.cmp1.performance_curve.head_isentropic[0]/1000,
        format="{:.2f} kJ/kg")
    new_tag(
        "gts1_head_isen",
        expr=m.fs.gts1.performance_curve.head_isentropic[0]/1000,
        format="{:.2f} kJ/kg")
    new_tag(
        "gts2_head_isen",
        expr=m.fs.gts2.performance_curve.head_isentropic[0]/1000,
        format="{:.2f} kJ/kg")
    new_tag(
        "gts3_head_isen",
        expr=m.fs.gts3.performance_curve.head_isentropic[0]/1000,
        format="{:.2f} kJ/kg")
    new_tag(
        "cmp1_eff_isen",
        expr=m.fs.cmp1.efficiency_isentropic[0]*100,
        format="{:.2f}%")
    new_tag(
        "gts1_eff_isen",
        expr=m.fs.gts1.efficiency_isentropic[0]*100,
        format="{:.2f}%")
    new_tag(
        "gts2_eff_isen",
        expr=m.fs.gts2.efficiency_isentropic[0]*100,
        format="{:.2f}%")
    new_tag(
        "gts3_eff_isen",
        expr=m.fs.gts3.efficiency_isentropic[0]*100,
        format="{:.2f}%")
    new_tag(
        "valve01_opening",
        expr=m.fs.valve01.valve_opening[0]*100,
        format="{:.2f}%")
    new_tag(
        "valve02_opening",
        expr=m.fs.valve02.valve_opening[0]*100,
        format="{:.2f}%")
    new_tag(
        "valve03_opening",
        expr=m.fs.valve03.valve_opening[0]*100,
        format="{:.2f}%")
    new_tag(
        "gt_total_power",
        expr=m.fs.gt_power[0]/1e6,
        format="{:.2f} MW")
    m.tags = tags
    m.tag_format = tag_format
    return tags, tag_format


def write_pfd_results(filename, tags, tag_format, infilename=None):
    """
    Write simulation results in a template PFD in svg format and save as
    filename.

    Args:
        filename: (str) file namd for output
        tags: (dict) tag keys and expression values
        tag_format: (dict) tag keys and format string values
        infilename: input file name, if you want to use an alternative diagram

    Returns:
        None
    """
    if infilename is None:
        infilename = os.path.join(this_file_dir(), "gas_turbine.svg")
    with open(infilename, "r") as f:
        iutil.svg_tag(
            svg=f,
            tags=tags,
            outfile=filename,
            tag_format=tag_format)


def performance_curves(m, flow_scale=0.896):
    """
    Add head and efficenty curves to the turbine stages in a model.

    Args:
        m: model with expcted structure

    Returns:
        None
    """
    # The flow scale variable less you scale the turbine up or down for differnt
    # nominal flows and differnt full load power outputs.
    m.fs.performace_flow_scale = pyo.Var(initialize=flow_scale)
    m.fs.performace_flow_scale.fix()
    fscale = m.fs.performace_flow_scale

    # PYLINT-TODO the names "eff_isen_eqn" and "head_isen_eqn" are reused below to define other constraint functions,
    # causing pylint to report function-redefined errors
    # this likely does not actually cause issues at runtime,
    # but it could be worth to check anyway if the pylint errors can be addressed
    # e.g. by giving unique names to each of the affected functions
    # pylint: disable=function-redefined

    # Efficiency curves for three stages
    @m.fs.gts1.performance_curve.Constraint(m.fs.time)
    def eff_isen_eqn(b, t):
        f = fscale*b.parent_block().control_volume.properties_in[t].flow_vol
        return b.parent_block().efficiency_isentropic[t] == 1.02*(
            1.4469E-14*f**5 -
            6.3333E-11*f**4 +
            6.6179E-08*f**3 -
            3.1728E-05*f**2 +
            7.7846E-03*f +
            1.0724E-01)
    @m.fs.gts2.performance_curve.Constraint(m.fs.time)
    def eff_isen_eqn(b, t):
        f = fscale*b.parent_block().control_volume.properties_in[t].flow_vol
        return b.parent_block().efficiency_isentropic[t] == 1.02*(
            2.6599E-16*f**5 -
            2.5894E-12*f**4 +
            6.0174E-09*f**3 -
            6.4156E-06*f**2 +
            3.5005E-03*f +
            1.0724E-01)
    @m.fs.gts3.performance_curve.Constraint(m.fs.time)
    def eff_isen_eqn(b, t):
        f = fscale*b.parent_block().control_volume.properties_in[t].flow_vol
        return b.parent_block().efficiency_isentropic[t] == 1.02*(
            5.8407E-18*f**5 -
            1.2203E-13*f**4 +
            6.0863E-10*f**3 -
            1.3927E-06*f**2 +
            1.6310E-03*f +
            1.0724E-01)
    # Head curves for three stages
    @m.fs.gts1.performance_curve.Constraint(m.fs.time)
    def head_isen_eqn(b, t):
        f = pyo.log(
            fscale*b.parent_block().control_volume.properties_in[t].flow_vol)
        return b.head_isentropic[t] == -(
            -2085.1*f**3 + 38433*f**2 - 150764*f + 422313)
    @m.fs.gts2.performance_curve.Constraint(m.fs.time)
    def head_isen_eqn(b, t):
        f = pyo.log(
            fscale*b.parent_block().control_volume.properties_in[t].flow_vol)
        return b.head_isentropic[t] == -(
            -1676.3*f**3 + 34916*f**2 - 173801*f + 456957)
    @m.fs.gts3.performance_curve.Constraint(m.fs.time)
    def head_isen_eqn(b, t):
        f = pyo.log(
            fscale*b.parent_block().control_volume.properties_in[t].flow_vol)
        return b.head_isentropic[t] == -(
            -1373.6*f**3 + 31759*f**2 - 188528*f + 500520)


def main(
    comps,
    rxns,
    phases,
    air_comp,
    ng_comp,
    m=None,
    initialize=True,
    flow_scale=0.896):
    """Generate and initialize the gas turbine flowsheet. This method returns
    a model and solver. The flow_scale argument can be used to scale the turnine
    up or down with the same relative performace.

    Args:
        comps: (set of str) a set of componets
        rxns: (dict) reactions with reaction name key and key component for
            conversion calculations
        phases: (list) phases potentially present
        air_comp: (dict) mole-fraction air composition
        ng_comp: (dict) mole-fraction natural gas composition
        m: (ConcreteModel) model to add flowsheet too.  If None create a new one.
        initialize: (bool) If true initialize the flowsheet
        flow_scale: (float) flow scale multiplier to scale the performace curves
            up or down. Original nominal volumetric flow/new nominal volumetric
            flow (default=0.896).

    Returns:
        (ConcreteModel) model, (Solver) solver
    """
    expand_arcs = pyo.TransformationFactory("network.expand_arcs")
    #
    # Model, flowsheet and properties
    #
    if m is None:
        m = pyo.ConcreteModel("Gas Turbine Model")
    if not hasattr(m, "fs"):
        m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.gas_prop_params = GenericParameterBlock(default=get_prop(comps, phases))
    m.fs.gas_prop_params.set_default_scaling("mole_frac_comp", 10)
    m.fs.gas_prop_params.set_default_scaling("mole_frac_phase_comp", 10)
    _mf_scale = {
        "Ar":100,
        "O2":100,
        "H2S":1000,
        "SO2":1000,
        "H2":1000,
        "CO":1000,
        "C2H4":1000,
        "C2H6":1000,
        "C3H8":1000,
        "C4H10":1000,
        "CO2":1000}
    for c, s in _mf_scale.items():
        m.fs.gas_prop_params.set_default_scaling(
            "mole_frac_comp", s, index=c)
        m.fs.gas_prop_params.set_default_scaling(
            "mole_frac_phase_comp", s, index=("Vap", c))
    m.fs.gas_prop_params.set_default_scaling(
        "enth_mol_phase", 1e-3, index="Vap")

    m.fs.gas_combustion = GenericReactionParameterBlock(
        default=get_rxn(m.fs.gas_prop_params, rxns))
    # Variable for mole-fraction of O2 in the flue gas.  To initialize the model
    # will use a fixed value here.
    m.fs.cmbout_o2_mol_frac = pyo.Var(m.fs.time, initialize=0.1157)
    m.fs.cmbout_o2_mol_frac.fix()
    #
    # Unit models
    #
    # inlet/outlet blocks
    m.fs.feed_air1 = um.Feed(default={"property_package": m.fs.gas_prop_params})
    m.fs.feed_fuel1 = um.Feed(default={"property_package": m.fs.gas_prop_params})
    m.fs.exhaust_1 = um.Product(default={"property_package": m.fs.gas_prop_params})
    # Valve roughly approximating variable stator vanes to control air flow.
    m.fs.vsv = um.Valve(default={
        "valve_function_callback":um.ValveFunctionType.linear,
        "property_package": m.fs.gas_prop_params})
    # Comprssor for air
    m.fs.cmp1 = um.Compressor(default={
        "property_package": m.fs.gas_prop_params,
        "support_isentropic_performance_curves":True})
    # Blade cooling air splitter
    m.fs.splt1 = um.Separator(default={
        "property_package": m.fs.gas_prop_params,
        "outlet_list":["air04", "air05", "air07", "air09"]})
    # Three valves for blade cooling air.
    m.fs.valve01 = um.Valve(default={
        "valve_function_callback":um.ValveFunctionType.linear,
        "property_package": m.fs.gas_prop_params})
    m.fs.valve02 = um.Valve(default={
        "valve_function_callback":um.ValveFunctionType.linear,
        "property_package": m.fs.gas_prop_params})
    m.fs.valve03 = um.Valve(default={
        "valve_function_callback":um.ValveFunctionType.linear,
        "property_package": m.fs.gas_prop_params})
    # Mixer for fuel injection. Since this is a steady state model the pressure
    # drop on the fuel control valve is lumped into this mixer
    m.fs.inject1 = um.Mixer(default={
        "property_package": m.fs.gas_prop_params,
        "inlet_list":["gas", "air"],
        "momentum_mixing_type":um.MomentumMixingType.none})
    # Three blade cooling air mixers. The blade cooling isn't explicitly modeled
    # so these mainly just maintain the proper mass and enegy balance.
    m.fs.mx1 = um.Mixer(default={
        "property_package": m.fs.gas_prop_params,
        "inlet_list":["gas", "air"],
        "momentum_mixing_type":um.MomentumMixingType.equality})
    m.fs.mx2 = um.Mixer(default={
        "property_package": m.fs.gas_prop_params,
        "inlet_list":["gas", "air"],
        "momentum_mixing_type":um.MomentumMixingType.equality})
    m.fs.mx3 = um.Mixer(default={
        "property_package": m.fs.gas_prop_params,
        "inlet_list":["gas", "air"],
        "momentum_mixing_type":um.MomentumMixingType.equality})
    # Combustor, for now assuming complete reactions and no NOx
    m.fs.cmb1 = um.StoichiometricReactor(default={
        "property_package": m.fs.gas_prop_params,
        "reaction_package": m.fs.gas_combustion,
        "has_pressure_change": True})
    # Three gas turbine stages, with performance curves for all three added by
    # the performance_curves() fuctions.
    m.fs.gts1 = um.Turbine(default={
        "property_package": m.fs.gas_prop_params,
        "support_isentropic_performance_curves":True})
    m.fs.gts2 = um.Turbine(default={
        "property_package": m.fs.gas_prop_params,
        "support_isentropic_performance_curves":True})
    m.fs.gts3 = um.Turbine(default={
        "property_package": m.fs.gas_prop_params,
        "support_isentropic_performance_curves":True})
    performance_curves(m)
    #
    # Additional constraints/expressions
    #
    # to make this simpler, just deal with VSV valve-approximation pressure drop
    # and forget the opening
    m.fs.vsv.pressure_flow_equation.deactivate()
    # O2 fraction in combustor constraint (to set var)
    @m.fs.Constraint(m.fs.time)
    def o2_flow(b, t):
        return m.fs.cmb1.control_volume.properties_out[t].mole_frac_comp["O2"] \
            == m.fs.cmbout_o2_mol_frac[t]
    # Fuel injector pressure is the air pressure. This lumps in the gas valve
    @m.fs.inject1.Constraint(m.fs.time)
    def mxpress_eqn(b, t):
        return b.mixed_state[t].pressure == b.air_state[t].pressure
    # The pressure drop in the cumbustor is just a fixed 5%.
    @m.fs.cmb1.Constraint(m.fs.time)
    def pressure_drop_eqn(b, t):
        return (0.95 ==
            b.control_volume.properties_out[t].pressure /
            b.control_volume.properties_in[t].pressure)
    # Constraints for complete combustion, use key compoents and 100% conversion
    @m.fs.cmb1.Constraint(m.fs.time, rxns.keys())
    def reaction_extent(b, t, r):
        k = rxns[r]
        prp = b.control_volume.properties_in[t]
        stc = -m.fs.gas_combustion.rate_reaction_stoichiometry[r, "Vap", k]
        extent = b.rate_reaction_extent[t, r]
        return extent == prp.flow_mol*prp.mole_frac_comp[k]/stc
    # Calculate the total gas turnine gross power output.
    @m.fs.Expression(m.fs.time)
    def gt_power_expr(b, t):
        return (
            m.fs.cmp1.control_volume.work[t] +
            m.fs.gts1.control_volume.work[t] +
            m.fs.gts2.control_volume.work[t] +
            m.fs.gts3.control_volume.work[t])
    # Add a varable and constraint for gross power.  This allows fixing power
    # for simulations where a specific power output is desired.
    m.fs.gt_power = pyo.Var(m.fs.time)
    @m.fs.Constraint(m.fs.time)
    def gt_power_eqn(b, t):
        return b.gt_power[t] == b.gt_power_expr[t]
    #
    # Arcs
    #
    # Arc names are imoprtant here becuase they match the PFD and are used for
    # tagging stream quantities
    m.fs.fuel01 = Arc(source=m.fs.feed_fuel1.outlet, destination=m.fs.inject1.gas)
    m.fs.air01 = Arc(source=m.fs.feed_air1.outlet, destination=m.fs.vsv.inlet)
    m.fs.air02 = Arc(source=m.fs.vsv.outlet, destination=m.fs.cmp1.inlet)
    m.fs.air03 = Arc(source=m.fs.cmp1.outlet, destination=m.fs.splt1.inlet)
    m.fs.air04 = Arc(source=m.fs.splt1.air04, destination=m.fs.inject1.air)
    m.fs.air05 = Arc(source=m.fs.splt1.air05, destination=m.fs.valve01.inlet)
    m.fs.air06 = Arc(source=m.fs.valve01.outlet, destination=m.fs.mx1.air)
    m.fs.air07 = Arc(source=m.fs.splt1.air07, destination=m.fs.valve02.inlet)
    m.fs.air08 = Arc(source=m.fs.valve02.outlet, destination=m.fs.mx2.air)
    m.fs.air09 = Arc(source=m.fs.splt1.air09, destination=m.fs.valve03.inlet)
    m.fs.air10 = Arc(source=m.fs.valve03.outlet, destination=m.fs.mx3.air)
    m.fs.g01 = Arc(source=m.fs.inject1.outlet, destination=m.fs.cmb1.inlet)
    m.fs.g02 = Arc(source=m.fs.cmb1.outlet, destination=m.fs.gts1.inlet)
    m.fs.g03 = Arc(source=m.fs.gts1.outlet, destination=m.fs.mx1.gas)
    m.fs.g04 = Arc(source=m.fs.mx1.outlet, destination=m.fs.gts2.inlet)
    m.fs.g05 = Arc(source=m.fs.gts2.outlet, destination=m.fs.mx2.gas)
    m.fs.g06 = Arc(source=m.fs.mx2.outlet, destination=m.fs.gts3.inlet)
    m.fs.g07 = Arc(source=m.fs.gts3.outlet, destination=m.fs.mx3.gas)
    m.fs.g08 = Arc(source=m.fs.mx3.outlet, destination=m.fs.exhaust_1.inlet)
    # Create arc constraints
    expand_arcs.apply_to(m)
    #
    # Tag model
    #
    tags, tag_format = tag_model(m)
    #
    # Set some scaling
    #
    for i in ["air05", "air07", "air09"]:
        iscale.set_scaling_factor(m.fs.splt1.split_fraction[0.0, i], 100)
    iscale.set_scaling_factor(m.fs.valve01.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.valve02.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.valve03.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.vsv.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.cmp1.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.gts1.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.gts2.control_volume.work, 1e-8)
    iscale.set_scaling_factor(m.fs.gts3.control_volume.work, 1e-8)
    iscale.set_scaling_factor(
        m.fs.gts1.control_volume.properties_in[0].flow_mol, 1e-5)
    iscale.set_scaling_factor(
        m.fs.gts2.control_volume.properties_in[0].flow_mol, 1e-5)
    iscale.set_scaling_factor(
        m.fs.gts3.control_volume.properties_in[0].flow_mol, 1e-5)
    iscale.set_scaling_factor(
        m.fs.gts1.control_volume.properties_out[0].flow_mol, 1e-5)
    iscale.set_scaling_factor(
        m.fs.gts2.control_volume.properties_out[0].flow_mol, 1e-5)
    iscale.set_scaling_factor(
        m.fs.gts3.control_volume.properties_out[0].flow_mol, 1e-5)
    for i, v in m.fs.cmb1.control_volume.rate_reaction_extent.items():
        if i[1] == "ch4_cmb":
            iscale.set_scaling_factor(v, 1e-2)
        else:
            iscale.set_scaling_factor(v, 1)
    for i, c in m.fs.cmb1.reaction_extent.items():
            iscale.constraint_scaling_transform(c, 1e-2)
    for i, v in m.fs.cmb1.control_volume.rate_reaction_generation.items():
        if i[2] in ["O2", "CO2", "CH4", "H2O"]:
            iscale.set_scaling_factor(v, 0.001)
        else:
            iscale.set_scaling_factor(v, 0.1)
    for v in m.fs.cmp1.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.gts1.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.gts2.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.gts3.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.valve01.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.valve02.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for v in m.fs.valve03.deltaP.values():
        iscale.set_scaling_factor(v, 1e-6)
    for c in m.fs.o2_flow.values():
        iscale.constraint_scaling_transform(c, 10)
    for c in m.fs.gt_power_eqn.values():
        iscale.constraint_scaling_transform(c, 1e-8)
    for c in m.fs.mx1.enthalpy_mixing_equations.values():
        iscale.constraint_scaling_transform(c, 1e-8)
    for c in m.fs.mx2.enthalpy_mixing_equations.values():
        iscale.constraint_scaling_transform(c, 1e-8)
    for c in m.fs.mx3.enthalpy_mixing_equations.values():
        iscale.constraint_scaling_transform(c, 1e-8)
    for c in m.fs.inject1.enthalpy_mixing_equations.values():
        iscale.constraint_scaling_transform(c, 1e-8)
    for c in m.fs.gts1.performance_curve.head_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 1e-5)
    for c in m.fs.gts1.performance_curve.eff_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 2)
    for c in m.fs.gts2.performance_curve.head_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 1e-5)
    for c in m.fs.gts2.performance_curve.eff_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 2)
    for c in m.fs.gts3.performance_curve.head_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 1e-5)
    for c in m.fs.gts3.performance_curve.eff_isen_eqn.values():
        iscale.constraint_scaling_transform(c, 2)
    for c in m.fs.inject1.mxpress_eqn.values():
        iscale.constraint_scaling_transform(c, 1e-5)
    #
    # Calculate scaling factors/scaling transform of constraints
    #
    iscale.calculate_scaling_factors(m)
    #
    # Set basic model inputs for initialization
    #
    # compressor
    m.fs.vsv.deltaP.fix(-100)
    m.fs.cmp1.efficiency_isentropic.fix(0.9)
    m.fs.cmp1.ratioP.fix(19.075)
    # blabe cooling air valves use expected flow to calculate valve flow
    # coefficients, so here set expected flow and after init the opening will
    # be the fixed var.
    m.fs.valve01.control_volume.properties_in[0].flow_mol.fix(2315)
    m.fs.valve02.control_volume.properties_in[0].flow_mol.fix(509)
    m.fs.valve03.control_volume.properties_in[0].flow_mol.fix(354)
    m.fs.valve01.valve_opening.unfix()
    m.fs.valve02.valve_opening.unfix()
    m.fs.valve03.valve_opening.unfix()
    # Feed streams
    # Air at compressor inlet
    m.fs.feed_air1.flow_mol[:] = 34875.9
    m.fs.feed_air1.temperature.fix(288.15)
    m.fs.feed_air1.pressure.fix(101047)
    for i, v in air_comp.items():
        m.fs.feed_air1.mole_frac_comp[:,i].fix(v)
    # Gas at fuel injection
    m.fs.feed_fuel1.flow_mol.fix(1348.8)
    m.fs.feed_fuel1.temperature.fix(310.928)
    m.fs.feed_fuel1.pressure.fix(3.10264e6)
    for i, v in ng_comp.items():
        m.fs.feed_fuel1.mole_frac_comp[:,i].fix(v)
    #
    # Initialization
    #
    solver = pyo.SolverFactory('ipopt')
    if initialize:
        # feeds
        m.fs.feed_air1.initialize()
        m.fs.feed_fuel1.initialize()
        # compressor
        iutil.copy_port_values(m.fs.air01)
        m.fs.vsv.initialize()
        iutil.copy_port_values(m.fs.air02)
        m.fs.cmp1.initialize()
        # splitter
        iutil.copy_port_values(m.fs.air03)
        m.fs.splt1.split_fraction[0, "air05"].fix(0.0916985*0.73)
        m.fs.splt1.split_fraction[0, "air07"].fix(0.0916985*0.27*0.59)
        m.fs.splt1.split_fraction[0, "air09"].fix(0.0916985*0.27*0.41)
        m.fs.splt1.initialize()
        m.fs.splt1.split_fraction[0, "air05"].unfix()
        m.fs.splt1.split_fraction[0, "air07"].unfix()
        m.fs.splt1.split_fraction[0, "air09"].unfix()
        # inject
        iutil.copy_port_values(m.fs.air04)
        iutil.copy_port_values(m.fs.fuel01)
        m.fs.inject1.mixed_state[0].pressure = pyo.value(m.fs.inject1.air.pressure[0])
        m.fs.inject1.initialize()
        # combustor
        iutil.copy_port_values(m.fs.g01)
        m.fs.cmb1.initialize()
        # gas turbine stage 1
        iutil.copy_port_values(m.fs.g02)
        m.fs.gts1.ratioP[0] = 0.7
        m.fs.gts1.initialize()
        # blade cooling air valve01, and calculate a flow coefficent
        iutil.copy_port_values(m.fs.air05)
        m.fs.valve01.Cv = 2
        m.fs.valve01.Cv.unfix()
        m.fs.valve01.valve_opening.fix(0.85)
        m.fs.valve01.control_volume.properties_out[0].pressure.fix(
            pyo.value(m.fs.gts1.control_volume.properties_out[0].pressure))
        m.fs.valve01.initialize()
        m.fs.valve01.control_volume.properties_out[0].pressure.unfix()
        m.fs.valve01.Cv.fix()
        m.fs.valve01.valve_opening.unfix()
        # mixer 1
        iutil.copy_port_values(m.fs.air06)
        iutil.copy_port_values(m.fs.g03)
        m.fs.mx1.mixed_state[0].pressure = pyo.value(m.fs.mx1.gas.pressure[0])
        m.fs.mx1.initialize()
        # gas turbine stage 2
        iutil.copy_port_values(m.fs.g04)
        m.fs.gts2.ratioP[0] = 0.7
        m.fs.gts2.initialize()
        # blade cooling air valve02, and calculate a flow coefficent
        iutil.copy_port_values(m.fs.air07)
        m.fs.valve02.Cv = 2
        m.fs.valve02.Cv.unfix()
        m.fs.valve02.valve_opening.fix(0.85)
        m.fs.valve02.control_volume.properties_out[0].pressure.fix(
            pyo.value(m.fs.gts2.control_volume.properties_out[0].pressure))
        m.fs.valve02.initialize()
        m.fs.valve02.control_volume.properties_out[0].pressure.unfix()
        m.fs.valve02.Cv.fix()
        m.fs.valve02.valve_opening.unfix()
        # mixer 2
        iutil.copy_port_values(m.fs.air08)
        iutil.copy_port_values(m.fs.g05)
        m.fs.mx2.mixed_state[0].pressure = pyo.value(m.fs.mx2.gas.pressure[0])
        m.fs.mx2.initialize()
        # gas turbine stage 3
        iutil.copy_port_values(m.fs.g06)
        m.fs.gts3.ratioP[0] = 0.7
        m.fs.gts3.initialize()
        # blade cooling air valve03, and calculate a flow coefficent
        iutil.copy_port_values(m.fs.air09)
        m.fs.valve03.Cv = 2
        m.fs.valve03.Cv.unfix()
        m.fs.valve03.valve_opening.fix(0.85)
        m.fs.valve03.control_volume.properties_out[0].pressure.fix(
            pyo.value(m.fs.gts3.control_volume.properties_out[0].pressure))
        m.fs.valve03.initialize()
        m.fs.valve03.control_volume.properties_out[0].pressure.unfix()
        m.fs.valve03.Cv.fix()
        m.fs.valve03.valve_opening.unfix()
        # mixer 3
        iutil.copy_port_values(m.fs.air10)
        iutil.copy_port_values(m.fs.g07)
        m.fs.mx3.mixed_state[0].pressure = pyo.value(m.fs.mx3.gas.pressure[0])
        m.fs.mx3.initialize()    # product blocks
        # product blocks
        iutil.copy_port_values(m.fs.g08)
        m.fs.exhaust_1.initialize()
        # Solve
        solver.solve(m, tee=True)
    return m, solver

def run_full_load(m, solver):
    """ Set the model up for off-design pressure driven flow.  Run 480 MW case.

    Args:
        m: (ConcreteModel) model to run
        solver: (Solver)

    Returns:
        None
    """
    m.fs.feed_fuel1.flow_mol.unfix()
    # For initialization, flow was specified and and valve opening calculated
    m.fs.valve01.control_volume.properties_in[0].flow_mol.unfix()
    m.fs.valve02.control_volume.properties_in[0].flow_mol.unfix()
    m.fs.valve03.control_volume.properties_in[0].flow_mol.unfix()
    # deltaP will be whatever is needed to satisfy the power requirment
    m.fs.vsv.deltaP.unfix()
    # The compressor efficiency is a little high since it doesn't include
    # throttling in the valve use to approximate VSV.
    m.fs.cmp1.efficiency_isentropic.fix(0.92)
    m.fs.cmp1.ratioP.fix(17.5) # lowering this ratio, just means less pressure
                               # drop in the VSV valve, decresing throttle loss
    # Exhaust pressure will be a bit over ATM due to HRSG. This will come from
    # HRSG model when coupled to form NGCC model
    m.fs.gts3.control_volume.properties_out[0].pressure.fix(102594)
    # Don't know how much blade cooling air is needed for off desing case, but
    # full load flows were based on WVU model.  For now just leave valves at
    # fixed opening.
    m.fs.valve01.valve_opening.fix()
    m.fs.valve02.valve_opening.fix()
    m.fs.valve03.valve_opening.fix()
    # Feeds
    # Air at compressor inlet
    m.fs.feed_air1.temperature.fix(288.15)
    m.fs.feed_air1.pressure.fix(103421)
    # Gas at fuel injection
    m.fs.feed_fuel1.temperature.fix(310.928)
    m.fs.feed_fuel1.pressure.fix(3.10264e6)
    # It looks like the O2 or air/fuel ratio is actually determined to roughly
    # get a constant exhaust temperature.  In a real gas turbine there are
    # probably a lot of complex factors, but here we fix the exhaust temp.
    m.fs.cmbout_o2_mol_frac.unfix()
    m.fs.exhaust_1.temperature.fix(910)
    # Fix the gross power output
    m.fs.gt_power[0].fix(-480e6) # Watts negative because is power out
    # Solve
    return solver.solve(m, tee=True)

def run_series(m, solver):
    """Run off design at 480 MW then a series of lower loads. Call run_full_load
    first to configure the model properly before calling this function. Saves
    PFD and row in tabulated csv file results for each case.

    Args:
        m: (ConcreteModel) model to run
        solver: (Solver)

    Returns:
        None
    """
    tags = m.tags
    tag_format = m.tag_format
    # Write the results to a CSV with these columns
    columns=[
        "power", "Fuel_flow", "Air_flow", "TI_temp", "TI_pressure",
        "TI_flow", "EX_temp", "EX_pressure", "EX_flow", "FG_O2", "FG_CO2"]

    # Write CSV header
    with open("res.csv", "w", newline='') as f:
        cw = csv.writer(f)
        cw.writerow(columns)

    # Run simulations, write CSV rows, and write flowsheets
    for i, p in enumerate([480 - 48*j for j in range(9)]):
        print(f"Loop {p}") # Print power to track progress
        m.fs.gt_power[0].fix(-p*1e6) # fix gross power
        solver.solve(m, tee=True)
        with open("res.csv", "a", newline='') as f:
            csv.writer(f).writerow([
                p,
                pyo.value(tags["fuel01_F"]), # mass flow
                pyo.value(tags["air01_F"]), # mass flow
                pyo.value(tags["g02_T"]),
                pyo.value(tags["g02_P"]),
                pyo.value(tags["g02_F"]),
                pyo.value(tags["g08_T"]),
                pyo.value(tags["g08_P"]),
                pyo.value(tags["g08_F"]),
                pyo.value(tags["g08_yO2"]),
                pyo.value(tags["g08_yCO2"])])
        write_pfd_results(f"gas_turbine_results_{p}.svg", tags, tag_format)


if __name__ == "__main__":
    #
    # Solver config
    #
    # Use idaes config, because these will apply to all ipopt solvers created
    use_idaes_solver_configuration_defaults()
    idaes.cfg.ipopt["options"]["nlp_scaling_method"] = "user-scaling"
    # due to a lot of component mole fractions being on their lower bound of 0
    # bound push result in much longer solve times, so set it low.
    idaes.cfg.ipopt["options"]["bound_push"] = 1e-10


    comps = { # components present
        "CH4", "C2H6", "C2H4", "CO", "H2S", "H2", "O2", "H2O", "CO2", "N2",
        "Ar", "SO2"}
    rxns = { # reactions and key compoents for conversion
        "ch4_cmb":"CH4",
        "c2h6_cmb":"C2H6",
        "c2h4_cmb":"C2H4",
        "co_cmb":"CO",
        "h2s_cmb":"H2S",
        "h2_cmb":"H2"}
    phases = ["Vap"]
    air_comp = {
        "CH4":0.0,
        "C2H6":0.0,
        "C2H4":0.0,
        "CO":0.0,
        "H2S":0.0,
        "H2":0.0,
        "O2":0.2074,
        "H2O":0.0099,
        "CO2":0.0003,
        "N2":0.7732,
        "Ar":0.0092,
        "SO2":0.0}
    ng_comp = {
        "CH4":0.87,
        "C2H6":0.0846,
        "C2H4":0.0003,
        "CO":0.0009,
        "H2S":0.0004,
        "H2":0.0036,
        "O2":0.0007,
        "H2O":0.0,
        "CO2":0.0034,
        "N2":0.0361,
        "Ar":0.0,
        "SO2":0.0}

    m, solver = main(
        comps=comps,
        rxns=rxns,
        phases=phases,
        air_comp=air_comp,
        ng_comp=ng_comp)
    run_full_load(m, solver)
    #iscale.constraint_autoscale_large_jac(m)
    jac, nlp = iscale.get_jacobian(m, scaled=True)
    print("Extreme Jacobian entries:")
    for i in iscale.extreme_jacobian_entries(jac=jac, nlp=nlp, large=100):
        print(f"    {i[0]:.2e}, [{i[1]}, {i[2]}]")
    print("Unscaled constraints:")
    for c in iscale.unscaled_constraints_generator(m):
        print(f"    {c}")
    print("Scaled constraints by factor:")
    for c, s in iscale.constraints_with_scale_factor_generator(m):
        print(f"    {c}, {s}")
    print("Badly scaled variables:")
    for v, sv in iscale.badly_scaled_var_generator(m, large=1e2, small=1e-2, zero=1e-12):
        print(f"    {v} -- {sv} -- {iscale.get_scaling_factor(v)}")
    print(f"Jacobian Condition Number: {iscale.jacobian_cond(jac=jac):.2e}")
    write_pfd_results("gas_turbine_results.svg", m.tags, m.tag_format)
    #run_series(m, solver)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for rate forms
"""

import pytest

from pyomo.environ import Block, ConcreteModel, Var, units as pyunits

from idaes.generic_models.properties.core.generic.generic_reaction import \
    GenericReactionParameterBlock, ConcentrationForm
from idaes.generic_models.properties.core.reactions.rate_forms import *

from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.misc import add_object_reference


@pytest.mark.unit
def test_power_law_rate_no_order():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    m.rparams = GenericReactionParameterBlock(default={
        "property_package": m.pparams,
        "base_units": {"time": pyunits.s,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "length": pyunits.m,
                       "temperature": pyunits.K},
        "rate_reactions": {
            "r1": {"stoichiometry": {("p1", "c1"): -1,
                                     ("p1", "c2"): 2},
                   "rate_form": power_law_rate,
                   "concentration_form": ConcentrationForm.moleFraction}}})

    # Create a dummy state block
    m.rxn = Block([1])
    add_object_reference(
        m.rxn[1], "phase_component_set", m.pparams._phase_component_set)
    add_object_reference(m.rxn[1], "params", m.rparams)
    add_object_reference(m.rxn[1], "state_ref", m.thermo[1])

    m.rxn[1].k_rxn = Var(["r1"], initialize=1)

    power_law_rate.build_parameters(
        m.rparams.reaction_r1, m.rparams.config.rate_reactions["r1"])

    # Check parameter construction
    assert isinstance(m.rparams.reaction_r1.reaction_order, Var)
    assert len(m.rparams.reaction_r1.reaction_order) == 4
    for i, v in m.rparams.reaction_r1.reaction_order.items():
        try:
            stoic = m.rparams.config.rate_reactions.r1.stoichiometry[i]
        except KeyError:
            stoic = 0

        if stoic < 1:
            assert v.value == -stoic
        else:
            assert v.value == 0

    # Check reaction form
    rform = power_law_rate.return_expression(
        m.rxn[1], m.rparams.reaction_r1, "r1", 300)

    assert str(rform) == str(
        m.rxn[1].k_rxn["r1"] *
        m.thermo[1].mole_frac_phase_comp["p1", "c1"] **
        m.rparams.reaction_r1.reaction_order["p1", "c1"])


@pytest.mark.unit
def test_power_law_rate_with_order():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    m.rparams = GenericReactionParameterBlock(default={
        "property_package": m.pparams,
        "base_units": {"time": pyunits.s,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "length": pyunits.m,
                       "temperature": pyunits.K},
        "rate_reactions": {
            "r1": {"stoichiometry": {("p1", "c1"): -1,
                                     ("p1", "c2"): 2},
                   "rate_form": power_law_rate,
                   "concentration_form": ConcentrationForm.moleFraction,
                   "parameter_data": {
                       "reaction_order": {("p1", "c1"): 1,
                                          ("p1", "c2"): 2,
                                          ("p2", "c1"): 3,
                                          ("p2", "c2"): 4}}}}})

    # Create a dummy state block
    m.rxn = Block([1])
    add_object_reference(
        m.rxn[1], "phase_component_set", m.pparams._phase_component_set)
    add_object_reference(m.rxn[1], "params", m.rparams)
    add_object_reference(m.rxn[1], "state_ref", m.thermo[1])

    m.rxn[1].k_rxn = Var(["r1"], initialize=1)

    power_law_rate.build_parameters(
        m.rparams.reaction_r1, m.rparams.config.rate_reactions["r1"])

    # Check parameter construction
    assert isinstance(m.rparams.reaction_r1.reaction_order, Var)
    assert len(m.rparams.reaction_r1.reaction_order) == 4
    assert m.rparams.reaction_r1.reaction_order["p1", "c1"].value == 1
    assert m.rparams.reaction_r1.reaction_order["p1", "c2"].value == 2
    assert m.rparams.reaction_r1.reaction_order["p2", "c1"].value == 3
    assert m.rparams.reaction_r1.reaction_order["p2", "c2"].value == 4

    # Check reaction form
    rform = power_law_rate.return_expression(
        m.rxn[1], m.rparams.reaction_r1, "r1", 300)

    assert str(rform) == str(
        m.rxn[1].k_rxn["r1"] * (
            m.thermo[1].mole_frac_phase_comp["p1", "c1"] **
            m.rparams.reaction_r1.reaction_order["p1", "c1"] *
            m.thermo[1].mole_frac_phase_comp["p1", "c2"] **
            m.rparams.reaction_r1.reaction_order["p1", "c2"] *
            m.thermo[1].mole_frac_phase_comp["p2", "c1"] **
            m.rparams.reaction_r1.reaction_order["p2", "c1"] *
            m.thermo[1].mole_frac_phase_comp["p2", "c2"] **
            m.rparams.reaction_r1.reaction_order["p2", "c2"]))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

"""
Tests for update_workshop_materials
"""
import idaes.util.update_workshop_materials as up
import pytest

@pytest.mark.unit
def test_update_workshop_materials():
    # Note: This tests that the methods in update_workshop_materials.py
    # successfully download install_idaes_workshop_materials.py from Pyomo.org
    # We purposely do NOT test code within that downloaded module since we do
    # not want to automatically execute code during testing that was downloaded
    # from outside the IDAES repository.
    
    # download the module from pyomo.org
    download_dest = up.download_install_module()

    # check that the file contains the desired function (execute())
    download_succeeded = False
    with open(download_dest, 'r') as fd:
        for line in fd:
            if line.startswith('def execute():'):
                download_succeeded = True
                break

    assert download_succeeded == True

if __name__ == '__main__':
    test_update_workshop_materials()
    



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""Commandline Utilities for Managing the IDAES Data Directory"""

__author__ = "John Eslick"

import os
import click
import logging
import idaes
import idaes.util.download_bin
from idaes.commands import cb

_log = logging.getLogger("idaes.commands.extensions")


def print_extensions_version(library_only=False):
    click.echo("---------------------------------------------------")
    click.echo("IDAES Extensions Build Versions")
    click.echo("===================================================")
    if not library_only:
        v = os.path.join(idaes.bin_directory, "version_solvers.txt")
        try:
            with open(v, "r") as f:
                v = f.readline().strip()
        except FileNotFoundError:
            v = "no version file found"
        click.echo("Solvers:  v{}".format(v))
    v = os.path.join(idaes.bin_directory, "version_lib.txt")
    try:
        with open(v, "r") as f:
            v = f.readline().strip()
    except FileNotFoundError:
        v = "no version file found"
    click.echo("Library:  v{}".format(v))
    click.echo("===================================================")
    return 0


@cb.command(name="get-extensions", help="Get solvers and libraries")
@click.option(
    "--release",
    help="Optional, specify an official binary release to download",
    default=None)
@click.option(
    "--url",
    help="Optional, URL to download solvers/libraries from",
    default=None)
@click.option(
    "--platform",
    help="Platform to download binaries for. Use 'idaes get-extensions-platforms'"
         " for a list of available platforms (default=auto)",
    default="auto")
@click.option(
    "--insecure",
    is_flag=True,
    help="Don't verify download location")
@click.option(
    "--cacert",
    help="Specify certificate file to verify download location",
    default=None)
@click.option(
    "--nochecksum",
    is_flag=True,
    help="Don't verify the file checksum")
@click.option(
    "--library-only",
    is_flag=True,
    help="Only install shared physical property function libraries, not solvers")
@click.option(
    "--no-download",
    is_flag=True,
    help="Don't download anything, but report what would be done")
@click.option(
    "--show-current-version",
    is_flag=True,
    help="Show the version information if any for the currently installed")
@click.option(
    "--show-platforms",
    is_flag=True,
    help="Show the platform options")
@click.option(
    "--show-extras",
    is_flag=True,
    help="Show list of binary extras")
@click.option(
    "--extra",
    multiple=True,
    help="Install extras")
@click.option(
    "--extras-only",
    is_flag=True,
    help="Only install extras")
@click.option(
    "--to",
    default=None,
    help="Put extensions in a alternate location")
@click.option("--verbose", help="Show details", is_flag=True)
def get_extensions(
    release,
    url,
    insecure,
    cacert,
    verbose,
    platform,
    nochecksum,
    library_only,
    no_download,
    show_current_version,
    show_platforms,
    show_extras,
    extras_only,
    extra,
    to):

    if show_platforms:
        click.echo("\nSupported platforms for IDAES binary extensions.")
        for key, mes in sorted(idaes.config.known_binary_platform.items()):
            click.echo(f"    {key}: {mes}")
        click.echo("\nBinaries compiled on these platforms:")
        for k in sorted(idaes.config.basic_platforms()):
            click.echo(f"    {k}")
        return
    elif show_extras:
        click.echo("\nBinary Extras")
        for k in sorted(idaes.config.extra_binaries):
            click.echo(f"    {k}")
        return
    elif show_current_version:
        print_extensions_version()
        return
    if url is None and release is None:
        # the default release is only used if neither a release or url is given
        release = idaes.config.default_binary_release
    if url is not None and release is not None:
        click.echo("\n* You must provide either a release or url not both.")
    elif url is not None or release is not None:
        click.echo("Getting files...")
        try:
            d = idaes.util.download_bin.download_binaries(
                release,
                url,
                insecure,
                cacert,
                verbose,
                platform,
                nochecksum,
                library_only,
                no_download,
                extras_only,
                extra,
                alt_path=to)
            click.echo("Done")
        except idaes.util.download_bin.UnsupportedPlatformError as e:
            click.echo("")
            click.echo(e)
            click.echo("")
            click.echo(
                "Specify one of the following platforms with --platform <os>:")
            for i in sorted(idaes.config.known_binary_platform):
                if i == platform:
                    # auto or linux specified, and it didn't work.
                    # will need to be more specific.
                    continue
                click.echo(f"  {i}")
            return
        if no_download:
            for k, i in d.items():
                click.echo(f"{k:14}: {i}")
        else:
            print_extensions_version(library_only)
    else:
        click.echo("\n* You must provide a download URL for IDAES binary files.")


@cb.command(name="hash-extensions", help="Calculate release hashes")
@click.option(
    "--release",
    help="Optional, specify an official binary release to download",
    default=None,
    required=True)
@click.option(
    "--path",
    help="Directory of release files",
    default=None,
    required=True)
def hash_extensions(release, path):
    hfile = f"sha256sum_{release}.txt"
    if path is not None:
        hfile = os.path.join(path, hfile)

    def _write_hash(fp, pack, plat):
        f = f"idaes-{pack}-{plat}-64.tar.gz"
        if path is not None:
            h = idaes.util.download_bin.hash_file_sha256(os.path.join(path, f))
        else:
            h = idaes.util.download_bin.hash_file_sha256(f)
        fp.write(h)
        fp.write("  ")
        fp.write(f)
        fp.write("\n")

    with open(hfile, "w") as f:
        for plat in idaes.config.basic_platforms():
            for pack in ["solvers", "lib"]:
                _write_hash(f, pack, plat)
        for plat in idaes.config.basic_platforms():
            for pack, sp in idaes.config.extra_binaries.items():
                if plat not in sp:
                    continue
                _write_hash(f, pack, plat)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

__author__ = "Oluwamayowa Amusat, John Siirola"

from pyomo.core.expr import current as EXPR, native_types
from pyomo.core.expr.numvalue import value
from pyomo.common.collections import ComponentMap

_numpy_available = True
try:
    import numpy

    _functionMap = {
        'exp': numpy.exp,
        'log': numpy.log,
        'log10': numpy.log10,
        'sin': numpy.sin,
        'asin': numpy.arcsin,
        'sinh': numpy.sinh,
        'asinh': numpy.arcsinh,
        'cos': numpy.cos,
        'acos': numpy.arccos,
        'cosh': numpy.cosh,
        'acosh': numpy.arccosh,
        'tan': numpy.tan,
        'atan': numpy.arctan,
        'tanh': numpy.tanh,
        'atanh': numpy.arctanh,
        'ceil': numpy.ceil,
        'floor': numpy.floor,
        'sqrt': numpy.sqrt,
    }
except ImportError:
    _numpy_available = False


class NumpyEvaluator(EXPR.StreamBasedExpressionVisitor):

    def __init__(self, object_map):
        super(NumpyEvaluator, self).__init__()
        self.object_map = object_map

    def exitNode(self, node, values):
        if node.__class__ is EXPR.UnaryFunctionExpression or \
           node.__class__ is EXPR.NPV_UnaryFunctionExpression:
            return _functionMap[node._name](values[0])
        if node.__class__ is EXPR.AbsExpression or \
           node.__class__ is EXPR.NPV_AbsExpression:
            return numpy.abs(values[0])
        return node._apply_operation(values)

    def beforeChild(self, node, child, child_idx):
        #
        # Don't replace native types
        #
        if type(child) in native_types:
            return False, child
        #
        # We will descend into all expressions...
        #
        if child.is_expression_type():
            return True, None
        #
        # Replace pyomo variables with numpy variables
        #
        if child in self.object_map:
            return False, self.object_map[child]
        #
        # Assume everything else is a constant...
        #
        return False, value(child)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Pipe model for water or steam, Liq or Vap phase must be provided (mixed phase
not supported).

main equations:

* Heat is given (zero if adiabatic)
* Calculate pressure change due to friction and gravity
* Assume enthalpy_in = enthalpy_out + heat

Created: April 2019 by Jinliang Ma
"""
# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)

from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.constants import Constants as const
from idaes.core.util import get_solver
import idaes.logger as idaeslog


# Additional import for the unit operation
from pyomo.environ import SolverFactory, value, Var, Reference

__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "2.0.0"


@declare_process_block_class("WaterPipe")
class WaterPipeData(UnitModelBlockData):
    """
    Water or steam pipe Unit Class
    """
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.componentPhase,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.enthalpyTotal,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("contraction_expansion_at_end", ConfigValue(
        default='None',
        domain=In(['None', 'contraction', 'expansion']),
        description='Any contraction or expansion at the end',
        doc='Define if pressure drop due to contraction'
        ' or expansion needs to be considered'))
    CONFIG.declare("water_phase", ConfigValue(
        default='Liq',
        domain=In(['Liq', 'Vap']),
        description='Water phase',
        doc='''Define water phase for property calls,
mixed phase not supported'''))

    def build(self):
        """
        Begin building model
        """
        # Call UnitModel.build to setup dynamics
        super(WaterPipeData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
                "dynamic": self.config.dynamic,
                "has_holdup": self.config.has_holdup,
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args})

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=True)

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != EnergyBalanceType.none):
            self.heat_duty = Reference(self.control_volume.heat)

        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != 'none'):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary
        """
        # Number of pipe
        self.number_of_pipes = Var(
                initialize=4,
                doc="Number of pipes")
        # Length of pipe
        self.length = Var(
                initialize=10.0,
                doc="Total length of the straight pipe")
        # Elevation change of pipe, elevation of outlet - elevation of inlet
        self.elevation_change = Var(
                initialize=10.0,
                doc="Change of elevation from inlet to outlet")
        # Inside diameter of pipe
        self.diameter = Var(
                initialize=0.6,
                doc="Inside diameter of pipe")

        if not self.config.contraction_expansion_at_end == "None":
            self.area_ratio = Var(
                initialize=1,
                doc="Cross section area ratio of exit end to pipe")

        # Volume constraint
        @self.Constraint(self.flowsheet().time,
                         doc="Total volume of all pipes")
        def volume_eqn(b, t):
            return b.volume[t] == 0.25 * const.pi * b.diameter**2 \
                * b.length * b.number_of_pipes

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.
        """
        phase = self.config.water_phase

        # Add performance variables
        # Velocity of fluid inside pipe
        self.velocity = Var(
                self.flowsheet().time,
                initialize=10.0,
                doc='Fluid velocity inside pipe')

        # Reynolds number
        self.N_Re = Var(
                self.flowsheet().time,
                initialize=10000.0,
                doc='Reynolds number')

        # Darcy friction factor
        self.friction_factor_darcy = Var(
                self.flowsheet().time,
                initialize=0.005,
                doc='Darcy friction factor')

        # Correction factor for pressure drop due to friction
        self.fcorrection_dp = Var(initialize=1.0,
                                  doc="Correction factor for pressure drop")

        # Pressure change due to friction
        self.deltaP_friction = Var(
            self.flowsheet().time,
            initialize=-1.0,
            doc='Pressure change due to friction')

        # Pressure change due to gravity
        self.deltaP_gravity = Var(
                self.flowsheet().time,
                initialize=0.0,
                doc='Pressure change due to gravity')

        # Pressure change due to area change at end
        self.deltaP_area_change = Var(
                self.flowsheet().time,
                initialize=0.0,
                doc='Pressure change due to area change')

        # Equation for calculating velocity
        @self.Constraint(self.flowsheet().time,
                         doc="Velocity of fluid inside pipe")
        def velocity_eqn(b, t):
            return b.velocity[t]*0.25*const.pi*b.diameter**2\
                * b.number_of_pipes == \
                b.control_volume.properties_in[t].flow_vol

        # Equation for calculating Reynolds number
        @self.Constraint(self.flowsheet().time, doc="Reynolds number")
        def Reynolds_number_eqn(b, t):
            return b.N_Re[t] * \
                   b.control_volume.properties_in[t].visc_d_phase[phase] == \
                   b.diameter * b.velocity[t] *\
                   b.control_volume.properties_in[t].dens_mass_phase[phase]

        # Friction factor expression depending on laminar or turbulent flow
        @self.Constraint(self.flowsheet().time,
                         doc="Darcy friction factor as"
                         " a function of Reynolds number")
        def friction_factor_darcy_eqn(b, t):
            return b.friction_factor_darcy[t] * b.N_Re[t]**(0.25) == \
                0.3164 * b.fcorrection_dp

        # Pressure change equation for friction
        @self.Constraint(self.flowsheet().time,
                         doc="Pressure change due to friction")
        def pressure_change_friction_eqn(b, t):
            return b.deltaP_friction[t] * b.diameter == \
                   - 0.5 * b.control_volume.properties_in[t].\
                       dens_mass_phase[phase] * \
                   b.velocity[t]**2 * b.friction_factor_darcy[t] * b.length

        # Pressure change equation for gravity
        @self.Constraint(self.flowsheet().time,
                         doc="Pressure change due to gravity")
        def pressure_change_gravity_eqn(b, t):
            return b.deltaP_gravity[t] == -const.acceleration_gravity * \
                b.control_volume.properties_in[t].dens_mass_phase[phase]\
                * b.elevation_change

        # Pressure change equation for contraction or expansion
        @self.Constraint(self.flowsheet().time,
                         doc="Pressure change due to gravity")
        def pressure_change_area_change_eqn(b, t):
            if self.config.contraction_expansion_at_end == "contraction":
                return b.deltaP_area_change[t] == - (0.1602*b.area_ratio**2
                                                     - 0.646*b.area_ratio
                                                     + 1.4858) * \
                    0.5 * b.control_volume.properties_out[t].\
                    dens_mass_phase[phase] * (b.velocity[t]/b.area_ratio)**2 \
                    + 0.5*b.control_volume.properties_out[t].\
                    dens_mass_phase[phase]*b.velocity[t]**2
            elif self.config.contraction_expansion_at_end == "expansion":
                return b.deltaP_area_change[t] == \
                    b.control_volume.properties_out[t].dens_mass_phase[phase]\
                    * b.velocity[t]**2*(b.area_ratio-1)/b.area_ratio**2
            else:
                return b.deltaP_area_change[t] == 0

        # Total pressure change equation
        @self.Constraint(self.flowsheet().time, doc="Pressure drop")
        def pressure_change_total_eqn(b, t):
            return b.deltaP[t] == (b.deltaP_friction[t]
                                   + b.deltaP_gravity[t]
                                   + b.deltaP_area_change[t])

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)

    def initialize(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        WaterPipe initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        '''
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags = blk.control_volume.initialize(
            outlvl=outlvl+1,
            optarg=optarg,
            solver=solver,
            state_args=state_args
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # Fix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[t].pressure)
            )
        blk.pressure_change_total_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(res))
            )

        # Unfix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.unfix()
        blk.pressure_change_total_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
                "Initialization Step 3 {}.".format(idaeslog.condition(res))
            )
        blk.control_volume.release_state(flags, outlvl)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Workspace classes and functions.
"""
# stdlib
import logging
import os
import re
from typing import List
from urllib.parse import urlparse
import uuid

# third-party
import jsonschema
import yaml

# local
from .errors import (
    ParseError,
    WorkspaceError,
    WorkspaceNotFoundError,
    WorkspaceConfNotFoundError,
    WorkspaceConfMissingField,
    WorkspaceCannotCreateError,
)
from .util import yaml_load

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

_log = logging.getLogger(__name__)

CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://idaes.org/dmf/workspace-config",
    "type": "object",
    "properties": {
        "_id": {"description": "Unique workspace identifier", "type": "string"},
        "htmldocs": {
            "description": "HTML documentation paths",
            "type": "array",
            "items": {
                "type": "string",
                "description": "directory containing Sphinx HTML docs",
            },
            "default": "https://idaes-pse.readthedocs.io/en/stable,{dmf_root}/docs/build/html"
        },
        "description": {
            "description": "A human-readable description of the workspace",
            "type": "string",
        },
        "name": {"description": "A short name for the workspace", "type": "string"},
    },
}


class Fields(object):
    """Workspace configuration fields.
    """

    DOC_HTML_PATH = "htmldocs"  # path to documentation html dir
    LOG_CONF = "logging"  # logging config


class Workspace(object):
    """DMF Workspace.

    In essence, a workspace is some information at the root of a directory
    tree, a database (currently file-based, so also in the directory tree)
    of *Resources*, and a set of files associated with these resources.

    **Workspace Configuration**

    When the DMF is initialized, the workspace is given as a path to a
    directory. In that directory is a special file named ``config.yaml``,
    that contains metadata about the workspace. The very existence of a
    file by that name is taken by the DMF code as an indication that the
    containing directory is a DMF workspace::

        /path/to/dmf: Root DMF directory
         |
         +- config.yaml: Configuration file
         +- resourcedb.json: Resource metadata "database" (uses TinyDB)
         +- files: Data files for all resources

    The configuration file is a `YAML`_ formatted file

    .. _YAML: http://www.yaml.org/

    The configuration file defines the following key/value pairs:

        _id
            Unique identifier for the workspace. This is auto-generated by
            the library, of course.
        name
            Short name for the workspace.
        description
            Possibly longer text describing the workspace.
        created
            Date at which the workspace was created, as string in the
            ISO8601 format.
        modified
            Date at which the workspace was last modified, as string in the
            ISO8601 format.
        htmldocs
            Full path to the location of the built (not source) Sphinx HTML
            documentation for the `idaes_dmf` package. See
            DMF Help Configuration for more details.

    There are many different possible "styles" of formatting a list of values
    in YAML, but we prefer the simple block-indented style, where the key is
    on its own line and the values are each indented with a dash:

    .. code-block:: YAML

        _id: fe5372a7e51d498fb377da49704874eb
        created: '2018-07-16 11:10:44'
        description: A bottomless trashcan
        modified: '2018-07-16 11:10:44'
        name: Oscar the Grouch's Home
        htmldocs:
        - '{dmf_root}/doc/build/html/dmf'
        - '{dmf_root}/doc/build/html/models'


    Any paths in the workspace configuration, e.g., for the "htmldocs",
    can use two special variables that will take on values relative to the
    workspace location. This avoids hardcoded paths and makes the workspace
    more portable across environments. ``{ws_root}`` will be replaces with
    the path to the workspace directory, and ``{dmf_root}`` will be replaced
    with the path to the (installed) DMF package.

    The `config.yaml` file will allow keys and values it does not know
    about. These will be accessible, loaded into a Python dictionary,
    via the ``meta`` attribute on the :class:`Workspace` instance.
    This may be useful for
    passing additional user-defined information into the DMF at startup.
    """

    #: Name of configuration file placed in WORKSPACE_DIR
    WORKSPACE_CONFIG = "config.yaml"
    #: Name of ID field
    ID_FIELD = "_id"

    CONF_NAME = "name"  #: Configuration field for name
    CONF_DESC = "description"  #: Configuration field for description
    CONF_CREATED = "created"  #: Configuration field for created date
    CONF_MODIFIED = "modified"  #: Configuration field for modified date

    def __init__(
        self, path, create=False, existing_ok=True, add_defaults=False, html_paths=None
    ):
        """Load or create a workspace rooted at `path`

        Args:
            (str) path: Path for root of workspace
            (bool) create: Create the workspace if it does not exist.
            (bool) existing_ok: If create is True, and the workspace exists, just continue
            (bool) add_defaults: Add default values to new configuration.
                    These values are found in the JSON Schema that is stored in the module variable
                    `CONFIG_SCHEMA`, and substituted by logic in :meth:`WorkspaceConfiguration.get_fields`.
            html_paths: One or more paths to HTML docs (or None)
        Raises:
            WorkspaceNotFoundError: if ws is not found (and create is false)
            WorkspaceConfNotFoundError: if ws config is not found (& ~create)
            WorkspaceConfMissingField: if there is no ID field.
            DMFError: Anything else
        """
        path = os.path.abspath(path)
        self._wsdir = path
        self._conf = os.path.join(self._wsdir, self.WORKSPACE_CONFIG)
        self._cached_conf = None
        if create:
            skip_config = False
            # note: these raise OSError on failure
            try:
                os.mkdir(self._wsdir, 0o770)
            except OSError:
                if not os.path.exists(self._wsdir):
                    raise WorkspaceCannotCreateError(self._wsdir)
                if os.path.exists(self._conf):
                    if existing_ok:
                        _log.info(f"Using existing DMF workspace: {self._conf}")
                        skip_config = True
                    else:
                        raise WorkspaceError(
                            "existing configuration would be "
                            "overwritten: {}".format(self._conf)
                        )
                _log.info(
                    "Using existing path for new DMF workspace: {}".format(self._wsdir)
                )
            if not skip_config:
                try:
                    self._create_new_config(add_defaults)
                except OSError as err:
                    raise WorkspaceError(
                        "while creating new workspace " "configuration: {}".format(err)
                    )
        else:
            # assert that the workspace exists
            try:
                assert os.path.isdir(self._wsdir)
            except AssertionError:
                raise WorkspaceNotFoundError(self._wsdir)
            try:
                assert os.path.isfile(self._conf)
            except AssertionError:
                raise WorkspaceConfNotFoundError(self._conf)
        self._install_dir = self._get_install_dir()
        try:
            self._id = self.meta[self.ID_FIELD]
        except KeyError:
            raise WorkspaceConfMissingField(path, self.ID_FIELD, "ID field")
        if html_paths:
            self.set_doc_paths(html_paths)

    def _create_new_config(self, add_defaults):
        _log.info(f"Create new configuration at '{self._conf}'")
        conf = open(self._conf, 'w')  # create the file
        new_id = uuid.uuid4().hex  # create new unique ID
        conf.write("{}: {}\n".format(self.ID_FIELD, new_id))  # write ID
        conf.close()  # flush and close
        if add_defaults:
            self._configure_defaults()

    def _configure_defaults(self):
        """Add default values to the workspace configuration.

        The default values are provided by the w

        Note: this will *overwrite* any existing values!
        """
        values = {
            key: val
            for key, (desc, val) in WorkspaceConfiguration()
            .get_fields(only_defaults=True)
            .items()
        }
        self.set_meta(values)

    @property
    def wsid(self):
        """Get workspace identifier (from config file).

        Returns:
            str: Unique identifier.
        """
        return self._id

    @staticmethod
    def _get_install_dir():
        fpath = os.path.realpath(__file__)
        fdir = os.path.dirname(fpath)
        idir = os.path.join(fdir, "..", "..")
        return os.path.abspath(idir)

    def set_meta(self, values, remove=None):
        # type: (dict, list) -> None
        """Update metadata with new values.
        
        Args:
            values (dict): Values to add or change
            remove (list): Keys of values to remove.
        """
        d = self._read_conf()
        if remove:
            for key in remove:
                if key in d:
                    del d[key]
                else:
                    _log.warning('Cannot remove "{}": no such key'.format(key))
        d.update(values)
        self._write_conf(d)
        self._cached_conf = d

    @property
    def meta(self):
        # type: () -> dict
        """Get metadata.

        This reads and parses the configuration.
        Therefore, one way to force a config refresh is to
        simply refer to this property, e.g.::

             dmf = DMF(path='my-workspace')
             #  ... do stuff that alters the config ...
             dmf.meta  # re-read/parse the config

        Returns:
            (dict) Metadata for this workspace.
        """
        # Re-read configuration from file
        d = self._read_conf()
        # In sections that contain paths,
        # substitute special variables in the values with
        # the paths for this workspace.
        if Fields.DOC_HTML_PATH in d:
            p = d[Fields.DOC_HTML_PATH]
            if isinstance(p, str):
                p = [p]
            d[Fields.DOC_HTML_PATH] = list(map(self._expand_path, p))
        if Fields.LOG_CONF in d:
            # Look through logging configuration tree, and
            # replace paths in any value whose key is 'output'.
            stack = [(k, d) for k in d]
            while stack:
                key, p = stack.pop()
                if key == "output":
                    p[key] = self._expand_path(p[key])
                elif isinstance(p[key], dict):
                    for key2 in p[key].keys():
                        stack.append((key2, p[key]))
        # Return modified configuration metadata
        return d

    def get_doc_paths(self):
        """Get paths to generated HTML Sphinx docs.

        Returns:
            (list) Paths or empty list if not found.
        """
        paths = self.meta.get(Fields.DOC_HTML_PATH, [])
        if len(paths) and hasattr(paths, "lower"):
            paths = [paths]  # make a str into a list
        return paths

    def set_doc_paths(self, paths: List[str], replace: bool = False):
        """Set paths to generated HTML Sphinx docs.

        Args:
            paths: New paths to add.
            replace: If True, replace any existing paths. Otherwise merge
                     new paths with existing ones.
        """
        existing_paths = self.get_doc_paths()
        if replace:
            new_paths = list(set(existing_paths + paths))
        else:
            new_paths = paths
        if new_paths != existing_paths:
            self.set_meta({Fields.DOC_HTML_PATH: new_paths})

    def _read_conf(self):
        #: type (None) -> dict
        if self._cached_conf is None:
            _log.debug('Load workspace configuration from "{}"'.format(self._conf))
            conf = open(self._conf, "r")
            try:
                contents = yaml_load(conf)
            except Exception as err:
                raise ParseError(
                    'Cannot load config file "{f}": {e}'.format(f=self._conf, e=err)
                )
            if contents is None:  # empty file
                contents = {}
            self._cached_conf = contents
        else:
            contents = self._cached_conf
        if isinstance(contents, str):
            raise ParseError("File contents cannot be simple str ({})".format(contents))
        else:
            return contents.copy()

    def _expand_path(self, path):
        # leave paths bracketed by _underscores_ as-is
        if re.match(r"^_.*_$", path):
            pass
        else:
            # See if this is a file or web URL.
            # Do substitution on file paths, but leave web URLs alone.
            parsed = urlparse(path)
            if (not parsed.netloc) and parsed.path:
                # Do substitution of '{ws_root}' and '{dmf_root}' in the path
                path = os.path.realpath(
                    path.format(ws_root=self.root, dmf_root=self._install_dir)
                )
        return path

    def _write_conf(self, contents):
        #: type (dict) -> None
        conf = open(self._conf, "w")
        yaml.dump(contents, conf, default_flow_style=False)
        conf.close()

    @property
    def root(self):
        """Root path for this workspace.
        This is the path containing the configuration file.
        """
        return self._wsdir

    @property
    def configuration_file(self):
        """Configuration file path.
        """
        return self._conf

    @property
    def name(self):
        return self.meta.get(self.CONF_NAME, "none")

    @property
    def description(self):
        return self.meta.get(self.CONF_DESC, "none")


class WorkspaceConfiguration(object):
    """Interface to the :data:`CONFIG_SCHEMA` JSON Schema that specifies the fields in the
       workspace configuration.
    """
    DEFAULTS = {"string": "", "number": 0, "boolean": False, "array": []}

    def __init__(self):
        self._schema = jsonschema.Draft4Validator(CONFIG_SCHEMA)

    def get_fields(self, only_defaults=False) -> dict:
        """Get the metadata fields that are in a workspace config[uration], with default values
        provided for each field.

        The default values come out of the configuration schema in :data:`CONFIG_SCHEMA`.
        Keys starting with a leading underscore, like '_id', are skipped.

        Args:
            only_defaults: Do not return fields that are not given a default value in the schema.

        Returns:
            Keys are field name, values are (field description, value). The 'value' is the default value.
            Its type is either a list, a number, boolean, or a string.
        """
        result = {}  # return value

        # Loop over all properties in the schema
        for key, item in CONFIG_SCHEMA["properties"].items():
            # Skip properties whose name starts with a leading underscore
            if key.startswith("_"):
                continue
            desc, type_ = item["description"], item["type"]
            # Coerce unknown types to string (but print a warning, as this is not expected)
            if type_ not in self.DEFAULTS:
                _log.warning(
                    'Unknown schema type "{}".' 'Using "string" instead'.format(type_)
                )
                type_ = "string"
            # Based on value type, extract default value
            default_value = None
            if type_ == "array":
                # Arrays need special processing
                item_type, item_desc = item["items"]["type"], item["items"]["description"]
                desc = "{}. Each item is a {}".format(desc, item_desc)
                type_ = item_type
                # Use default either (1) inside the list of items, or (2) outside for the whole property
                if "default" in item["items"]:
                    default_value = [item["items"]["default"]]
                elif "default" in item:
                    # default is comma-separated list
                    default_value = item["default"].split(",")
            elif "default" in item:
                # For everything else, simply get the "default" key
                default_value = item["default"]
            # Set result, except if we are in 'only_defaults' mode and there is no default value
            if only_defaults:
                if default_value is None:
                    _log.debug(f"(only_defaults mode) NOT setting default value for {key}: no value provided")
                else:
                    # Set the value
                    result[key] = (desc, default_value)
                    _log.debug(f"(only_defaults mode) setting default value for {key} to: {default_value}")
            else:
                # If not in only_defaults mode, use an internal default value based on the type when there
                # is no default value in the schema.
                if default_value is None:
                    default_value = self.DEFAULTS[type_]
                # Set the value
                result[key] = (desc, default_value)
                _log.debug(f"setting default value for {key} to: {default_value}")

        return result


def find_workspaces(root):
    """Find workspaces at or below 'root'.

    Args:
        root (str): Path to start at

    Returns:
        List[str]: paths, which are all workspace roots.
    """
    w = []
    # try all subdirectories of root
    for (dirpath, dirnames, filenames) in os.walk(root):
        for dmfdir in dirnames:
            dmfpath = os.path.join(dirpath, dmfdir)
            conf_file = os.path.join(dmfpath, Workspace.WORKSPACE_CONFIG)
            if os.path.exists(conf_file):
                w.append(dmfpath)
    # also try root itself
    if os.path.exists(os.path.join(root, Workspace.WORKSPACE_CONFIG)):
        w.insert(0, root)
    return w



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
# matopt.materials.parsers

# TODO: Decide if (and how) we should support:
# from matopt.materials.parsers import *
# NOTE: Probably will never need this, since parsers should
#       probably just be incorporated in Design.py and most
#       of these methods should be behind-the-scenes.

'''
import matopt.materials.parsers.CFG
import matopt.materials.parsers.PDB
import matopt.materials.parsers.POSCAR
import matopt.materials.parsers.XYZ
'''
# TODO: Decide if this is necessary



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Created on Thu Mar 18 15:05:41 2021

@author: andrew
"""
import pytest

import pyomo.environ as pe
import idaes.core
import idaes.generic_models.unit_models
import idaes.generic_models.properties.swco2
from pyomo.network import Arc


@pytest.fixture(scope="function")
def model():
    m = pe.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock()
    m.fs.properties = \
        idaes.generic_models.properties.swco2.SWCO2ParameterBlock()
    m.fs.heater = idaes.generic_models.unit_models.Heater(default={
        'dynamic': False,
        'property_package': m.fs.properties,
        'has_pressure_change': True})
    m.fs.heater2 = idaes.generic_models.unit_models.Heater(default={
        'dynamic': False,
        'property_package': m.fs.properties,
        'has_pressure_change': True})
    m.fs.stream = Arc(source=m.fs.heater.outlet,
                      destination=m.fs.heater2.inlet)

    return m


@pytest.mark.unit
def test_expand_arcs_and_clone(model):
    # Check that port references were attached to model
    assert hasattr(model.fs.heater, "_enth_mol_inlet_ref")
    assert hasattr(model.fs.heater, "_flow_mol_inlet_ref")
    assert hasattr(model.fs.heater, "_pressure_inlet_ref")
    assert hasattr(model.fs.heater, "_enth_mol_outlet_ref")
    assert hasattr(model.fs.heater, "_flow_mol_outlet_ref")
    assert hasattr(model.fs.heater, "_pressure_outlet_ref")

    assert hasattr(model.fs.heater2, "_enth_mol_inlet_ref")
    assert hasattr(model.fs.heater2, "_flow_mol_inlet_ref")
    assert hasattr(model.fs.heater2, "_pressure_inlet_ref")
    assert hasattr(model.fs.heater2, "_enth_mol_outlet_ref")
    assert hasattr(model.fs.heater2, "_flow_mol_outlet_ref")
    assert hasattr(model.fs.heater2, "_pressure_outlet_ref")

    assert model.fs.stream._expanded_block is None

    m2 = pe.TransformationFactory('network.expand_arcs').create_using(model)

    # Check that Arcs were expanded
    assert isinstance(m2.fs.stream._expanded_block, pe.Block)
    assert model.fs.stream._expanded_block is None


@pytest.mark.unit
def test_clone(model):
    m2 = model.clone()
    assert (m2.fs.heater.inlet.flow_mol[0] is
            m2.fs.heater.control_volume.properties_in[0].flow_mol)

    assert not (m2.fs.heater.inlet.flow_mol[0] is
                model.fs.heater.control_volume.properties_in[0].flow_mol)

    pe.TransformationFactory('network.expand_arcs').apply_to(m2)

    # Check that Arcs were expanded
    assert isinstance(m2.fs.stream._expanded_block, pe.Block)
    assert model.fs.stream._expanded_block is None



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
NRTL property model for a benzene-toluene mixture.
The example model is from the IDAES tutorial,
https://github.com/IDAES/examples-pse/blob/main/src/Tutorials/Advanced/ParamEst/
"""
from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models import Flash
from idaes.generic_models.properties.activity_coeff_models.BTX_activity_coeff_VLE import BTXParameterBlock
import idaes.logger as idaeslog
from pyomo.environ import *

def NRTL_model(data):
    """This function generates an instance of the NRTL Pyomo model using 'data' as the input argument
    
    Parameters
    ----------
    data: pandas DataFrame, list of dictionaries, or list of json file names
        Data that is used to build an instance of the Pyomo model
    
    Returns
    -------
    m: an instance of the Pyomo model
        for estimating parameters and covariance
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                 ('Liq', 'Vap'),
                                                 "activity_coeff_model":
                                                 'NRTL'})
    m.fs.flash = Flash(default={"property_package": m.fs.properties})

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368) 
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Initialize the flash unit
    m.fs.flash.initialize(outlvl=idaeslog.INFO_LOW)

    # Fix at actual temperature
    m.fs.flash.inlet.temperature.fix(float(data["temperature"]))

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)

    # Return initialized flash model
    return m


def NRTL_model_opt():
    """This function generates an instance of the NRTL Pyomo model
    
    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                 ('Liq', 'Vap'),
                                                 "activity_coeff_model":
                                                 'NRTL'})
    m.fs.flash = Flash(default={"property_package": m.fs.properties})

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Initialize the flash unit
    m.fs.flash.initialize(outlvl=idaeslog.INFO_LOW)

    # Fix at actual temperature
    m.fs.flash.inlet.temperature.fix(float(368))

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)
    
    # To use kaug
    # objective function required
    # need to unfix the variables
    m.obj = Objective(expr = 0*m.fs.properties.tau["benzene","toluene"] + exp(-m.fs.properties.alpha['toluene','benzene'].value * m.fs.properties.tau['toluene','benzene']), sense=minimize)
    m.fs.properties.tau["benzene", "toluene"].fixed = False # To use kaug
    m.fs.properties.tau["toluene", "benzene"].fixed = False
    return m


def NRTL_model_opt_infeasible():
    """This function generates an instance of the NRTL Pyomo model
    
    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                 ('Liq', 'Vap'),
                                                 "activity_coeff_model":
                                                 'NRTL'})
    m.fs.flash = Flash(default={"property_package": m.fs.properties})

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)

    # To use kaug
    # objective function required
    # need to unfix the variables
    m.obj = Objective(expr = 0*m.fs.properties.tau["benzene","toluene"] + exp(-m.fs.properties.alpha['toluene','benzene'].value * m.fs.properties.tau['toluene','benzene']), sense=minimize)
    m.fs.properties.tau["benzene", "toluene"].fixed = False
    m.fs.properties.tau["toluene", "benzene"].fixed = False
    return m



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Calculates dimensionless data
"""

import math

molecule = ""
critT, critD, critP, acc, R, M, Rm = [0, 0, 0, 0, 0, 0, 0]


def molData(fluidData, mol, RVal):
    """
    Sets up important global values
    """
    global critT, critP, critD, M, triple, acc, R, Rm, molecule
    (critT, critP, critD, M, triple, acc) = fluidData
    molecule = mol
    R = RVal
    Rm = RVal


def PVT(x):
    """
    Calculate dimensionless compressibility

    Args:
        X: [Pressure, Density, Temperature]

    Returns:
        [Delta, Tau, Compressibility]
    """
    Pressure = float(x[0])
    Density = float(x[1])
    Temperature = float(x[2])

    Tau = float(critT) / Temperature
    Delta = Density / float(critD)

    if Density != 0:
        Z = (Pressure * 1000 / R / Temperature / Density) - 1.00
    else:
        Z = 0
    return [Delta, Tau, Z]


def P(x):
    """
    Calculate reduced density and inverse reduced temperature
    Return array of Delta, Tau, Pressure
    """
    Temperature = float(x[2])
    Density = float(x[1])
    Pressure = float(x[0])
    Tau = float(critT) / Temperature
    Delta = Density / float(critD)
    return [Delta, Tau, Pressure]


def CP(x):
    """
    Calculate dimensionless isobaric heat capacity

    Args:
        X: [Density, Temperature, Isobaric Heat Capacity]

    Returns:
        [Delta, Tau, CP]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    CP = float(x[2]) / R

    return [Delta, Tau, CP]


def CV(x):
    """
    Calculate dimensionless isochoric heat capacity

    Args:
        X: [Density, Temperature, Isochoric Heat Capacity]

    Returns:
        [Delta, Tau, CV]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    CV = float(x[2]) / R
    return [Delta, Tau, CV]


def SND(x):
    """
    Calculate dimensionless speed of sound

    Args:
        X: [Density, Temperature, Speed of Sound]

    Returns:
        [Delta, Tau, W]
    """
    Delta = float(x[0]) / float(critD)
    Tau = float(critT) / float(x[1])
    w = (float(x[2]) ** 2) / R / 1000 * M / float(x[1])
    return [Delta, Tau, w]


def CP0(x): 
    """
    Calculate dimensionless ideal isobaric heat capacity
    """
    Cp0 = float(x[1]) / Rm / 1000.0
    Temp = float(x[0]) / 1000.0
    return [Temp, Cp0]


def DL(x):
    """
    Calculate Theta and Delta for saturated liquid density

    Args:
         X: [Density, Temperature]

    Returns:
        [Theta, Delta]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Delta = float(x[0]) / float(critD) - 1
    return [Theta, Delta]


def DV(x):
    """
    Calculate Theta and Delta for saturated vapor density

    Args:
         X: [Density, Temperature]

    Returns:
        [Theta, Delta]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Delta = math.log(float(x[0]) / float(critD))
    return [Theta, Delta]


def Dsat(x):
    """
    Calculate dimensionless terms
    """
    Theta = float(critT) / float(x[1])
    Delta = float(x[0]) / float(critD)
    return [Delta, Theta]


def PV(x):
    """
    Calculate Theta, Tau, and Psi for saturated liquid density

    Args:
         X: [Pressure, Temperature]

    Returns:
        [Tau, Theta, Psi]
    """
    Theta = 1 - float(x[1]) / float(critT)
    Psi = math.log(float(x[0]) / float(critP))
    Tau = float(critT) / float(x[1])
    return [Tau, Theta, Psi]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for setting up FpcTP as the state variables in a generic property
package
"""
from pyomo.environ import \
    Constraint, Expression, NonNegativeReals, Var, value, units as pyunits

from idaes.core import (MaterialFlowBasis,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.generic_models.properties.core.generic.utility import \
    get_bounds_from_config
from .electrolyte_states import \
    define_electrolyte_state, calculate_electrolyte_scaling
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

# Set up logger
_log = idaeslog.getLogger(__name__)


def set_metadata(b):
    # The default metadata should be correct in this case, so no need to update
    pass


def define_state(b):
    # FpcTP contains full information on the phase equilibrium, so flash
    # calculations re not always needed
    b.always_flash = False

    # Check that only necessary state_bounds are defined
    expected_keys = ["flow_mol_phase_comp", "enth_mol",
                     "temperature", "pressure"]
    if (b.params.config.state_bounds is not None and
            any(b.params.config.state_bounds.keys()) not in expected_keys):
        for k in b.params.config.state_bounds.keys():
            if k not in expected_keys:
                raise ConfigurationError(
                    "{} - found unexpected state_bounds key {}. Please ensure "
                    "bounds are provided only for expected state variables "
                    "and that you have typed the variable names correctly."
                    .format(b.name, k))

    units = b.params.get_metadata().derived_units
    # Get bounds and initial values from config args
    f_bounds, f_init = get_bounds_from_config(
        b, "flow_mol_phase_comp", units["flow_mole"])
    t_bounds, t_init = get_bounds_from_config(
        b, "temperature", units["temperature"])
    p_bounds, p_init = get_bounds_from_config(
        b, "pressure", units["pressure"])

    # Add state variables
    b.flow_mol_phase_comp = Var(b.phase_component_set,
                                initialize=f_init,
                                domain=NonNegativeReals,
                                bounds=f_bounds,
                                doc='Phase-component molar flowrate',
                                units=units["flow_mole"])
    b.pressure = Var(initialize=p_init,
                     domain=NonNegativeReals,
                     bounds=p_bounds,
                     doc='State pressure',
                     units=units["pressure"])
    b.temperature = Var(initialize=t_init,
                        domain=NonNegativeReals,
                        bounds=t_bounds,
                        doc='State temperature',
                        units=units["temperature"])

    # Add supporting variables
    b.flow_mol = Expression(
        expr=sum(b.flow_mol_phase_comp[i]
                 for i in b.phase_component_set),
        doc="Total molar flowrate")

    def flow_mol_phase(b, p):
        return sum(b.flow_mol_phase_comp[p, j]
                   for j in b.component_list
                   if (p, j) in b.phase_component_set)
    b.flow_mol_phase = Expression(b.phase_list,
                                  rule=flow_mol_phase,
                                  doc='Phase molar flow rates')

    def rule_flow_mol_comp(b, j):
        return sum(b.flow_mol_phase_comp[p, j]
                   for p in b.phase_list
                   if (p, j) in b.phase_component_set)
    b.flow_mol_comp = Expression(b.component_list,
                                 rule=rule_flow_mol_comp,
                                 doc='Component molar flow rates')

    def mole_frac_comp(b, j):
        return (sum(b.flow_mol_phase_comp[p, j]
                    for p in b.phase_list
                    if (p, j) in b.phase_component_set) / b.flow_mol)
    b.mole_frac_comp = Expression(b.component_list,
                                  rule=mole_frac_comp,
                                  doc='Mixture mole fractions')

    b.mole_frac_phase_comp = Var(
            b.phase_component_set,
            initialize=1/len(b.component_list),
            doc='Phase mole fractions',
            units=None)

    def rule_mole_frac_phase_comp(b, p, j):
        # Calcualting mole frac phase comp is degenerate if there is only one
        # component in phase.
        # Count components
        comp_count = 0
        for p1, j1 in b.phase_component_set:
            if p1 == p:
                comp_count += 1

        if comp_count > 1:
            return b.mole_frac_phase_comp[p, j] * b.flow_mol_phase[p] == \
                b.flow_mol_phase_comp[p, j]
        else:
            return b.mole_frac_phase_comp[p, j] == 1
    b.mole_frac_phase_comp_eq = Constraint(
        b.phase_component_set, rule=rule_mole_frac_phase_comp)

    def rule_phase_frac(b, p):
        if len(b.phase_list) == 1:
            return 1
        else:
            return b.flow_mol_phase[p] / b.flow_mol
    b.phase_frac = Expression(
        b.phase_list,
        rule=rule_phase_frac,
        doc='Phase fractions')

    # Add electrolye state vars if required
    if b.params._electrolyte:
        define_electrolyte_state(b)

    # -------------------------------------------------------------------------
    # General Methods
    def get_material_flow_terms_FpcTP(p, j):
        """Create material flow terms for control volume."""
        return b.flow_mol_phase_comp[p, j]
    b.get_material_flow_terms = get_material_flow_terms_FpcTP

    def get_enthalpy_flow_terms_FpcTP(p):
        """Create enthalpy flow terms."""
        # enth_mol_phase probably does not exist when this is created
        # Use try/except to build flow term if not present
        try:
            eflow = b._enthalpy_flow_term
        except AttributeError:
            def rule_eflow(b, p):
                return b.flow_mol_phase[p] * b.enth_mol_phase[p]
            eflow = b._enthalpy_flow_term = Expression(
                b.phase_list, rule=rule_eflow)
        return eflow[p]
    b.get_enthalpy_flow_terms = get_enthalpy_flow_terms_FpcTP

    def get_material_density_terms_FpcTP(p, j):
        """Create material density terms."""
        # dens_mol_phase probably does not exist when this is created
        # Use try/except to build term if not present
        try:
            mdens = b._material_density_term
        except AttributeError:
            def rule_mdens(b, p, j):
                return b.dens_mol_phase[p] * b.mole_frac_phase_comp[p, j]
            mdens = b._material_density_term = Expression(
                b.phase_component_set, rule=rule_mdens)
        return mdens[p, j]
    b.get_material_density_terms = get_material_density_terms_FpcTP

    def get_energy_density_terms_FpcTP(p):
        """Create energy density terms."""
        # Density and energy terms probably do not exist when this is created
        # Use try/except to build term if not present
        try:
            edens = b._energy_density_term
        except AttributeError:
            def rule_edens(b, p):
                return b.dens_mol_phase[p] * b.energy_internal_mol_phase[p]
            edens = b._energy_density_term = Expression(
                b.phase_list, rule=rule_edens)
        return edens[p]
    b.get_energy_density_terms = get_energy_density_terms_FpcTP

    def default_material_balance_type_FpcTP():
        return MaterialBalanceType.componentTotal
    b.default_material_balance_type = default_material_balance_type_FpcTP

    def default_energy_balance_type_FpcTP():
        return EnergyBalanceType.enthalpyTotal
    b.default_energy_balance_type = default_energy_balance_type_FpcTP

    def get_material_flow_basis_FpcTP():
        return MaterialFlowBasis.molar
    b.get_material_flow_basis = get_material_flow_basis_FpcTP

    def define_state_vars_FpcTP():
        """Define state vars."""
        return {"flow_mol_phase_comp": b.flow_mol_phase_comp,
                "temperature": b.temperature,
                "pressure": b.pressure}
    b.define_state_vars = define_state_vars_FpcTP

    def define_display_vars_FpcTP():
        """Define display vars."""
        return {"Molar Flowrate": b.flow_mol_phase_comp,
                "Temperature": b.temperature,
                "Pressure": b.pressure}
    b.define_display_vars = define_display_vars_FpcTP


def state_initialization(b):
    for i in b.phase_component_set:
        b.mole_frac_phase_comp[i].value = value(
            b.flow_mol_phase_comp[i] / b.flow_mol_phase[i[0]])


def define_default_scaling_factors(b):
    """
    Method to set default scaling factors for the property package. Scaling
    factors are based on the default initial value for each variable provided
    in the state_bounds config argument.
    """
    # Get bounds and initial values from config args
    units = b.get_metadata().derived_units
    state_bounds = b.config.state_bounds

    if state_bounds is None:
        return

    try:
        f_bounds = state_bounds["flow_mol_phase_comp"]
        if len(f_bounds) == 4:
            f_init = pyunits.convert_value(f_bounds[1],
                                           from_units=f_bounds[3],
                                           to_units=units["flow_mole"])
        else:
            f_init = f_bounds[1]
    except KeyError:
        f_init = 1

    try:
        p_bounds = state_bounds["pressure"]
        if len(p_bounds) == 4:
            p_init = pyunits.convert_value(p_bounds[1],
                                           from_units=p_bounds[3],
                                           to_units=units["pressure"])
        else:
            p_init = p_bounds[1]
    except KeyError:
        p_init = 1

    try:
        t_bounds = state_bounds["temperature"]
        if len(t_bounds) == 4:
            t_init = pyunits.convert_value(t_bounds[1],
                                           from_units=t_bounds[3],
                                           to_units=units["temperature"])
        else:
            t_init = t_bounds[1]
    except KeyError:
        t_init = 1

    # Set default scaling factors
    b.set_default_scaling("flow_mol", 1/f_init)
    b.set_default_scaling("flow_mol_phase", 1/f_init)
    b.set_default_scaling("flow_mol_comp", 1/f_init)
    b.set_default_scaling("flow_mol_phase_comp", 1/f_init)
    b.set_default_scaling("pressure", 1/p_init)
    b.set_default_scaling("temperature", 1/t_init)


def calculate_scaling_factors(b):
    for p, j in b.phase_component_set:
        sf = iscale.get_scaling_factor(b.flow_mol_phase_comp[
            p, j], default=1, warning=True)
        iscale.constraint_scaling_transform(
            b.mole_frac_phase_comp_eq[p, j], sf, overwrite=False)

    if b.params._electrolyte:
        calculate_electrolyte_scaling(b)


do_not_initialize = []


class FpcTP(object):
    set_metadata = set_metadata
    define_state = define_state
    state_initialization = state_initialization
    do_not_initialize = do_not_initialize
    define_default_scaling_factors = define_default_scaling_factors
    calculate_scaling_factors = calculate_scaling_factors



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for calculating rate constants
"""
from pyomo.environ import exp, Var, units as pyunits

from idaes.core import MaterialFlowBasis
from idaes.generic_models.properties.core.generic.generic_reaction import \
    ConcentrationForm
from idaes.core.util.misc import set_param_from_config
from idaes.core.util.constants import Constants as c
from idaes.core.util.exceptions import BurntToast, ConfigurationError


# -----------------------------------------------------------------------------
# Constant dh_rxn
class arrhenius():

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        rbasis = parent.config.reaction_basis
        if rbasis == MaterialFlowBasis.molar:
            r_base = units["amount"]
        elif rbasis == MaterialFlowBasis.mass:
            r_base = units["mass"]
        else:
            raise BurntToast(
                "{} for unexpected reaction basis {}. This should not happen "
                "so please contact the IDAES developers with this bug."
                .format(rblock.name, rbasis))

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name))
        elif (c_form == ConcentrationForm.moleFraction or
              c_form == ConcentrationForm.massFraction or
              c_form == ConcentrationForm.activity):
            r_units = r_base*units["volume"]**-1*units["time"]**-1
        else:
            order = 0
            for p, j in parent.config.property_package._phase_component_set:
                order += -rblock.reaction_order[p, j].value

            if c_form == ConcentrationForm.molarity:
                c_units = units["density_mole"]
            elif c_form == ConcentrationForm.molality:
                c_units = units["amount"]*units["mass"]**-1
            elif c_form == ConcentrationForm.partialPressure:
                c_units = units["pressure"]
            else:
                raise BurntToast(
                    "{} get_concentration_term received unrecognised "
                    "ConcentrationForm ({}). This should not happen - please "
                    "contact the IDAES developers with this bug."
                    .format(rblock.name, c_form))

            r_units = (r_base *
                       units["length"]**-3 *
                       units["time"]**-1 *
                       c_units**order)

        rblock.arrhenius_const = Var(
                doc="Arrhenius constant (pre-exponential factor)",
                units=r_units)
        set_param_from_config(rblock, param="arrhenius_const", config=config)

        rblock.energy_activation = Var(
                doc="Activation energy",
                units=units["energy_mole"])

        set_param_from_config(rblock, param="energy_activation", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        units = rblock.parent_block().get_metadata().derived_units

        return rblock.arrhenius_const * exp(
            -rblock.energy_activation / (
                pyunits.convert(c.gas_constant,
                                to_units=units["gas_constant"])*T))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""Commandline Utilities for Managing the IDAES Config files"""

__author__ = "John Eslick"

import os
import idaes
import click
import json
import pyomo.common.config
from idaes.commands import cb
import idaes.config as cfg

@cb.command(name="config-write", help="Write the IDAES config")
@click.option("--file", default="idaes_config.json", help="File name to write.")
@click.option(
    "--default", is_flag=True, help="Write the default config")
def config_write(file, default):
    idaes.write_config(file, default)
    click.echo(f"Wrote config to {file}")

@cb.command(name="config-file", help="Show the config file paths")
@click.option("--global", "gbl", is_flag=True, help="Global config path")
@click.option("--local", is_flag=True, help="Local config path")
def config_file(gbl, local):
    if gbl:
        click.echo(idaes._global_config_file)
    elif local:
        click.echo(idaes._local_config_file)
    else:
        click.echo(f"global: {idaes._global_config_file}")
        click.echo(f"local: {idaes._local_config_file}")

@cb.command(name="config-set", help="Set a set a configuration option")
@click.argument("key")
@click.argument("value")
@click.option("--global", "glb", is_flag=True, help="Write to global config")
@click.option("--local", is_flag=True, help="Write to local config")
@click.option("--file", help="Write to specified file", type=str)
@click.option("--display", is_flag=True, help="Show new config")
@click.option("--add", is_flag=True, help="Add to list or set")
@click.option("--del", "dlt", is_flag=True, help="Delete from to list or set")
@click.option(
    "--file_as_global",
    is_flag=True,
    help="Testing option, alternate global config location")
@click.option(
    "--file_as_local",
    is_flag=True,
    help="Testing option, alternate local config file location")
def config_set(
    glb, local, key, value, file, file_as_local, file_as_global, display, add, dlt):
    # get locations for the local and global config files
    global_config_file = idaes._global_config_file
    local_config_file = idaes._local_config_file
    # for testing allow an alternate location to be used for local and global
    # config files, to avoid messing with the real config files
    if file_as_local:
        local_config_file = file
        file = None
    if file_as_global:
        global_config_file = file
        file = None
    # make sure one and only one place is specified to write to
    if not (glb ^ local ^ bool(file)):
        click.echo("Must specify exactly one of --global, --local, --file")
        return
    # if global make sure you don't pick up local config and write it back
    if glb:
        idaes.cfg = cfg._new_idaes_config_block()
        idaes.read_config(global_config_file)
        idaes.cfg.display()
    elif local:
        idaes.cfg = cfg._new_idaes_config_block()
        idaes.read_config(global_config_file)
        idaes.read_config(local_config_file)
    idaes.cfg.display()
    # get the config block entry
    key = key.split(":") # use : instead of . due to logger names
    value = value.replace("'", '"') # " expected by json but gets eaten by shell
    value = json.loads(value)
    c = idaes.cfg
    for k in key[:-1]:
        try:
            c = c[k]
        except KeyError:
            if isinstance(c, pyomo.common.config.ConfigBlock):
                c[k] = pyomo.common.config.ConfigBlock(implicit=True)
            else:
                c[k] = {}
            c = c[k]
    # Set the config value
    if add and isinstance(c[key[-1]], list):
        c[key[-1]].append(value)
    elif add and isinstance(c[key[-1]], set):
        c[key[-1]].add(value)
    elif dlt and isinstance(c[key[-1]], (set, list)):
        c[key[-1]].remove(value)
    elif dlt or add:
        click.echo("--add or --del options must be used on a list or set")
    else:
        c[key[-1]] = value
    # Write changes
    if glb:
        idaes.write_config(global_config_file)
    elif local:
        idaes.write_config(local_config_file)
    elif file:
        idaes.write_config(file)
    if display:
        idaes.cfg.display()

@cb.command(name="config-display", help="Show IDAES config")
def config_display():
    idaes.cfg.display()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Base class for unit models
"""

from pyomo.environ import Reference
from pyomo.network import Port
from pyomo.common.config import ConfigValue, In

from .process_base import (declare_process_block_class,
                           ProcessBlockData,
                           useDefault)
from .property_base import StateBlock
from .control_volume_base import (ControlVolumeBlockData,
                                  FlowDirection,
                                  MaterialBalanceType)
from idaes.core.util.exceptions import (BurntToast,
                                        ConfigurationError,
                                        PropertyPackageError,
                                        BalanceTypeNotSupportedError)
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.core.util.unit_costing
import idaes.logger as idaeslog
from idaes.core.util import get_solver

__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ['UnitModelBlockData', 'UnitModelBlock']

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("UnitModelBlock")
class UnitModelBlockData(ProcessBlockData):
    """
    This is the class for process unit operations models. These are models that
    would generally appear in a process flowsheet or superstructure.
    """
    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("dynamic", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Dynamic model flag",
        doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}"""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Holdup construction flag",
        doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - construct holdup terms,
**False** - do not construct holdup terms}"""))

    def build(self):
        """
        General build method for UnitModelBlockData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super(UnitModelBlockData, self).build()

        # Set up dynamic flag and time domain
        self._setup_dynamics()

    def model_check(blk):
        """
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume and tries to call the model_check method of the
        controlVolume block. If an AttributeError is raised, the check is
        passed.

        More complex models should overload this method with a model_check
        suited to the particular application, especially if there are multiple
        ControlVolume blocks present.

        Args:
            None

        Returns:
            None
        """
        # Run control volume block model checks
        try:
            blk.controlVolume.model_check()
        except AttributeError:
            pass

    def add_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build Port objects in a unit model and
        connect these to a specified StateBlock.

        Keyword Args:
            name : name to use for Port object.
            block : an instance of a StateBlock to use as the source to
                    populate the Port object
            doc : doc string for Port object

        Returns:
            A Pyomo Port object and associated components.
        """
        # Validate block object
        if not isinstance(block, StateBlock):
            raise ConfigurationError("{} block object provided to add_port "
                                     "method is not an instance of a "
                                     "StateBlock object. IDAES port objects "
                                     "should only be associated with "
                                     "StateBlocks.".format(blk.name))

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        p._state_block = (block, )

        # Get dict of Port members and names
        member_list = block[
                blk.flowsheet().time.first()].define_port_members()

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                slicer = block[:].component(member_list[s].local_name)
            else:
                slicer = block[:].component(member_list[s].local_name)[...]

            r = Reference(slicer)
            setattr(blk, "_"+s+"_"+name+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def add_inlet_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build inlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name : name to use for Port object (default = "inlet").
            block : an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the inlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc : doc string for Port object (default = "Inlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                        "{} add_inlet_port was called without a block argument"
                        " but a name argument was provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))
            else:
                name = "inlet"
            # Try for default ControlVolume name
            try:
                block = blk.control_volume
            except AttributeError:
                raise ConfigurationError(
                        "{} add_inlet_port was called without a block argument"
                        " but no default ControlVolume exists "
                        "(control_volume). Please provide block to which the "
                        "Port should be associated.".format(blk.name))
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                        "{} add_inlet_port was called with a block argument, "
                        "but a name argument was not provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))

        if doc is None:
            doc = "Inlet Port"

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        # Get dict of Port members and names
        if isinstance(block, ControlVolumeBlockData):
            try:
                member_list = (block.properties_in[
                                    block.flowsheet().time.first()]
                               .define_port_members())
                p._state_block = (block.properties_in, )
            except AttributeError:
                try:
                    member_list = (block.properties[
                                    block.flowsheet().time.first(), 0]
                                   .define_port_members())
                    if block._flow_direction == FlowDirection.forward:
                        p._state_block = (block.properties,
                                          block.length_domain.first())
                    elif block._flow_direction == FlowDirection.backward:
                        p._state_block = (block.properties,
                                          block.length_domain.last())
                except AttributeError:
                    raise PropertyPackageError(
                            "{} property package does not appear to have "
                            "implemented a define_port_memebers method. "
                            "Please contact the developer of the property "
                            "package.".format(blk.name))
        elif isinstance(block, StateBlock):
            member_list = block[
                    blk.flowsheet().time.first()].define_port_members()
            p._state_block = (block, )
        else:
            raise ConfigurationError(
                    "{} block provided to add_inlet_port "
                    "method was not an instance of a "
                    "ControlVolume or a StateBlock."
                    .format(blk.name))

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_in[:].component(
                                        member_list[s].local_name)
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.first()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.last()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx]
                                  .component(member_list[s].local_name))
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))
            else:
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_in[:].component(
                                        member_list[s].local_name)[...]
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.first()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.last()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx].component(
                                    member_list[s].local_name))[...]
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)[...]
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))

            r = Reference(slicer)
            setattr(blk, "_"+s+"_"+name+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def add_outlet_port(blk, name=None, block=None, doc=None):
        """
        This is a method to build outlet Port objects in a unit model and
        connect these to a specified control volume or state block.

        The name and block arguments are optional, but must be used together.
        i.e. either both arguments are provided or neither.

        Keyword Args:
            name : name to use for Port object (default = "outlet").
            block : an instance of a ControlVolume or StateBlock to use as the
                    source to populate the Port object. If a ControlVolume is
                    provided, the method will use the outlet state block as
                    defined by the ControlVolume. If not provided, method will
                    attempt to default to an object named control_volume.
            doc : doc string for Port object (default = "Outlet Port")

        Returns:
            A Pyomo Port object and associated components.
        """
        if block is None:
            # Check that name is None
            if name is not None:
                raise ConfigurationError(
                        "{} add_outlet_port was called without a block "
                        "argument  but a name argument was provided. Either "
                        "both a name and a block must be provided or neither."
                        .format(blk.name))
            else:
                name = "outlet"
            # Try for default ControlVolume name
            try:
                block = blk.control_volume
            except AttributeError:
                raise ConfigurationError(
                        "{} add_outlet_port was called without a block "
                        "argument but no default ControlVolume exists "
                        "(control_volume). Please provide block to which the "
                        "Port should be associated.".format(blk.name))
        else:
            # Check that name is not None
            if name is None:
                raise ConfigurationError(
                        "{} add_outlet_port was called with a block argument, "
                        "but a name argument was not provided. Either both "
                        "a name and a block must be provided or neither."
                        .format(blk.name))

        if doc is None:
            doc = "Outlet Port"

        # Create empty Port
        p = Port(noruleinit=True, doc=doc)
        setattr(blk, name, p)

        # Get dict of Port members and names
        if isinstance(block, ControlVolumeBlockData):
            try:
                member_list = (block.properties_out[
                                    block.flowsheet().time.first()]
                               .define_port_members())
                p._state_block = (block.properties_out, )
            except AttributeError:
                try:
                    member_list = (block.properties[
                                    block.flowsheet().time.first(), 0]
                                   .define_port_members())
                    if block._flow_direction == FlowDirection.forward:
                        p._state_block = (block.properties,
                                          block.length_domain.last())
                    elif block._flow_direction == FlowDirection.backward:
                        p._state_block = (block.properties,
                                          block.length_domain.first())
                except AttributeError:
                    raise PropertyPackageError(
                            "{} property package does not appear to have "
                            "implemented a define_port_members method. "
                            "Please contact the developer of the property "
                            "package.".format(blk.name))
        elif isinstance(block, StateBlock):
            member_list = block[
                    blk.flowsheet().time.first()].define_port_members()
            p._state_block = (block, )
        else:
            raise ConfigurationError(
                    "{} block provided to add_inlet_port "
                    "method was not an instance of a "
                    "ControlVolume or a StateBlock."
                    .format(blk.name))

        # Create References for port members
        for s in member_list:
            if not member_list[s].is_indexed():
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_out[:].component(
                                        member_list[s].local_name)
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.last()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.first()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx]
                                  .component(member_list[s].local_name))
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))
            else:
                # Need to use slice notation on indexed comenent as well
                if isinstance(block, ControlVolumeBlockData):
                    try:
                        slicer = block.properties_out[:].component(
                                        member_list[s].local_name)[...]
                    except AttributeError:
                        if block._flow_direction == FlowDirection.forward:
                            _idx = block.length_domain.last()
                        elif block._flow_direction == FlowDirection.backward:
                            _idx = block.length_domain.first()
                        else:
                            raise BurntToast(
                                    "{} flow_direction argument received "
                                    "invalid value. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(blk.name))
                        slicer = (block.properties[:, _idx].component(
                                    member_list[s].local_name))[...]
                elif isinstance(block, StateBlock):
                    slicer = block[:].component(member_list[s].local_name)[...]
                else:
                    raise ConfigurationError(
                            "{} block provided to add_inlet_port "
                            "method was not an instance of a "
                            "ControlVolume or a StateBlock."
                            .format(blk.name))

            r = Reference(slicer)
            setattr(blk, "_"+s+"_"+name+"_ref", r)

            # Add Reference to Port
            p.add(r, s)

        return p

    def add_state_material_balances(self, balance_type, state_1, state_2):
        """
        Method to add material balances linking two State Blocks in a Unit
        Model. This method is not intended to replace Control Volumes, but
        to automate writing material balances linking isolated State Blocks
        in those models where this is required.

        Args:
            balance_type - a MaterialBalanceType Enum indicating the type
                            of material balances to write
            state_1 - first State Block to be linked by balances
            state_2 - second State Block to be linked by balances

        Returns:
            None
        """
        # Confirm that both state blocks stem from the same parameter block
        if not isinstance(state_1, StateBlock):
            raise ConfigurationError(
                    "{} state_1 argument to add_state_material_balances "
                    "was not an instance of a State Block.".format(self.name))

        if not isinstance(state_2, StateBlock):
            raise ConfigurationError(
                    "{} state_2 argument to add_state_material_balances "
                    "was not an instance of a State Block.".format(self.name))

        # Check that no constraint with the same name exists
        # We will only support using this method once per Block
        if hasattr(self, "state_material_balances"):
            raise AttributeError(
                    "{} a set of constraints named state_material_balances "
                    "already exists in the current UnitModel. To avoid "
                    "confusion, add_state_material_balances is only supported "
                    "once per UnitModel.".format(self.name))

        # Get a representative time point for testing
        rep_time = self.flowsheet().time.first()
        if state_1[rep_time].params is not state_2[rep_time].params:
            raise ConfigurationError(
                    "{} add_state_material_balances method was provided with "
                    "State Blocks are not linked to the same "
                    "instance of a Physical Parameter Block. This method "
                    "only supports linking State Blocks from the same "
                    "Physical Parameter Block.".format(self.name))

        if balance_type == MaterialBalanceType.useDefault:
            balance_type = (
                state_1[rep_time].default_material_balance_type()
            )

        phase_list = state_1.phase_list
        component_list = state_1.component_list
        pc_set = state_1.phase_component_set

        if balance_type == MaterialBalanceType.componentPhase:
            # TODO : Should we include an optional phase equilibrium term here
            # to allow for systems where a phase-transition may occur?

            @self.Constraint(
                self.flowsheet().time,
                pc_set,
                doc="State material balances",
            )
            def state_material_balances(b, t, p, j):
                return state_1[t].get_material_flow_terms(
                    p, j
                ) == state_2[t].get_material_flow_terms(p, j)

        elif balance_type == MaterialBalanceType.componentTotal:

            @self.Constraint(
                self.flowsheet().time,
                component_list,
                doc="State material balances",
            )
            def state_material_balances(b, t, j):
                return sum(
                    state_1[t].get_material_flow_terms(p, j)
                    for p in phase_list if (p, j) in pc_set
                ) == sum(
                    state_2[t].get_material_flow_terms(p, j)
                    for p in phase_list if (p, j) in pc_set
                )

        elif balance_type == MaterialBalanceType.total:

            @self.Constraint(
                self.flowsheet().time,
                doc="State material balances",
            )
            def state_material_balances(b, t):
                return sum(
                    state_1[t].get_material_flow_terms(p, j) for p, j in pc_set
                ) == sum(
                    state_2[t].get_material_flow_terms(p, j) for p, j in pc_set
                )

        elif balance_type == MaterialBalanceType.elementTotal:
            raise BalanceTypeNotSupportedError(
                "{} add_state_material_balances does not support "
                "MaterialBalanceType.elementTotal.".format(self.name)
            )
        elif balance_type == MaterialBalanceType.none:
            raise BalanceTypeNotSupportedError(
                "{} add_state_material_balances does not support "
                "MaterialBalanceType.None.".format(self.name)
            )
        else:
            raise BurntToast(
                "{} add_state_material_balances received an unexpected "
                "argument for balance_type. This should never happen. Please "
                "contact the IDAES developers with this bug.".format(self.name)
            )

    def _get_stream_table_contents(self, time_point=0):
        """
        Assume unit has standard configuration of 1 inlet and 1 outlet.

        Developers should overload this as appropriate.
        """
        try:
            return create_stream_table_dataframe({"Inlet": self.inlet,
                                                  "Outlet": self.outlet},
                                                 time_point=time_point)
        except AttributeError:
            raise ConfigurationError(
                    f"Unit model {self.name} does not have the standard Port "
                    f"names (inet and outlet). Please contact the unit model "
                    f"developer to develop a unit specific stream table.")

    def initialize(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        This is a general purpose initialization routine for simple unit
        models. This method assumes a single ControlVolume block called
        controlVolume, and first initializes this and then attempts to solve
        the entire unit.

        More complex models should overload this method with their own
        initialization routines,

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default IDAES solver)

        Returns:
            None
        '''
        if optarg is None:
            optarg = {}

        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize control volume block
        flags = blk.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        init_log.info_high('Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        # Solve unit

        # if costing block exists, deactivate
        if hasattr(blk, "costing"):
            blk.costing.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)

        init_log.info_high(
            "Initialization Step 2 {}.".format(idaeslog.condition(results))
        )

        # if costing block exists, activate and initialize
        if hasattr(blk, "costing"):
            blk.costing.activate()
            idaes.core.util.unit_costing.initialize(blk.costing)
        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.control_volume.release_state(flags, outlvl)

        init_log.info('Initialization Complete: {}'
                      .format(idaeslog.condition(results)))
